# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
import logging

from django.apps import apps
from django.conf import settings

logger = logging.getLogger(__name__)


class Module(abc.ABC):
    icon = "fa-solid fa-question"

    @property
    @abc.abstractmethod
    def name(self):
        pass

    @property
    @abc.abstractmethod
    def module_id(self):
        pass

    @abc.abstractmethod
    def get_url(self, project):
        pass

    @abc.abstractmethod
    def is_configured(self, project):
        pass


class ModuleSection:
    def __init__(self, name, project, is_configured_cache):
        self.name = name
        self.project = project
        self.modules = []
        self.is_configured_cache = is_configured_cache

    def is_configured(self, module):
        if module.name not in self.is_configured_cache:
            self.is_configured_cache[module.name] = module.is_configured(self.project)
        return self.is_configured_cache[module.name]

    def add(self, module):
        if module is not None:
            self.modules.append(module)

    @property
    def included(self):
        for module in self.modules:
            if self.is_configured(module):
                module.url = module.get_url(self.project)
                yield module

    @property
    def included_list(self):
        return list(self.included)

    @property
    def addable(self):
        module_index = get_module_index()
        configured_modules = {
            name for name, module in module_index.items() if self.is_configured(module)
        }
        for module in self.modules:
            fulfilled = all(
                module in configured_modules
                for module in settings.HELIPORT_DEPENDENCIES.get(module.module_id, [])
            )
            if not self.is_configured(module) and fulfilled:
                module.url = module.get_url(self.project)
                yield module

    @property
    def addable_list(self):
        return list(self.addable)


class PreconfiguredModuleSection(ModuleSection):
    def is_configured(self, module):
        return True


interface_warning_apps = set()
heliport_system_apps = set(settings.HELIPORT_SYSTEM_APPS)


def get_heliport_apps():
    for name, app in apps.app_configs.items():
        if hasattr(app.module, "interface"):
            yield name, app.module.interface
        elif (
            app.name in heliport_system_apps and app.name not in interface_warning_apps
        ):
            logger.warning(
                f'App "{app.name}" has no interface. '
                'Did you forget "from . import interface" in __init__.py?'
            )
            interface_warning_apps.add(app.name)


is_first_index = True


def get_module_index():
    index = {}

    for _, interface in get_heliport_apps():
        for module in dir(interface):
            attr = getattr(interface, module)
            if isinstance(attr, type) and attr != Module:
                instance = getattr(interface, module)()
                if isinstance(instance, Module):
                    index[instance.module_id] = instance

    global is_first_index
    if is_first_index:
        is_first_index = False
        logger.debug(f"installed modules: {list(index)}")

    return index


def generate_graph(project):
    module_index = get_module_index()
    graph = []
    is_configured_cache = {}

    for section_name, module_names in settings.HELIPORT_GRAPH.items():
        section = ModuleSection(section_name, project, is_configured_cache)
        for module_name in module_names:
            section.add(module_index.get(module_name))

        graph.append(section)

    return graph


def generate_project_dropdown(project):
    module_index = get_module_index()
    is_configured_cache = {}

    modules = PreconfiguredModuleSection("", project, is_configured_cache)
    for module_name in settings.HELIPORT_PROJECT_DROPDOWN_ITEMS:
        modules.add(module_index.get(module_name))

    return modules


def generate_project_navbar(project):
    module_index = get_module_index()
    is_configured_cache = {}

    modules = PreconfiguredModuleSection("", project, is_configured_cache)
    for module_name in settings.HELIPORT_PROJECT_NAVBAR_ITEMS:
        modules.add(module_index.get(module_name))

    return modules


def get_api_roots():
    results = []
    for _, interface in get_heliport_apps():
        if hasattr(interface, "get_api_root"):
            results.append(interface.get_api_root())
    return results


def get_search_urls():
    results = []
    for _, interface in get_heliport_apps():
        if hasattr(interface, "get_search_url"):
            results.append(interface.get_search_url())
    return results


def get_project_create_urls():
    results = {}
    for _, interface in get_heliport_apps():
        if hasattr(interface, "get_project_create_urls"):
            results.update(interface.get_project_create_urls())
    return results


def serialize_project_data(project):
    values = []
    objects = []
    lists = []
    for _, interface in get_heliport_apps():
        if hasattr(interface, "serialize_project"):
            app_result = interface.serialize_project(project)
            for key, value in app_result.items():
                if isinstance(value, list):
                    lists.append((key, value))
                elif isinstance(value, dict):
                    objects.append((key, value))
                else:
                    values.append((key, value))

    result = {
        "@context": {
            "datacite": "http://purl.org/spar/datacite/",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "heliport": "https://heliport/schema/",
            "time": "http://www.w3.org/2006/time#",
            "dc": "http://purl.org/dc/terms/",
        }
    }

    def add_to_results(k, v):
        if k in result:
            logger.error(f"colliding project serialization keys: {k}")
        result[k] = v

    for key, value in values:
        add_to_results(key, value)
    for key, value in objects:
        add_to_results(key, value)
    for key, value in lists:
        add_to_results(key, value)

    return result
