# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf
#
# SPDX-License-Identifier: GPL-3.0-or-later

import environ
from appconf import AppConf
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

env = environ.Env()


class HeliportAppConf(AppConf):
    """General settings of the :mod:`heliport.core` app.

    All settings in this class can be overwritten in settings.py
    """

    MAINTENANCE_MESSAGE_LEVEL = env.int(
        "HELIPORT_CORE_MAINTENANCE_MESSAGE_LEVEL", default=0
    )
    MAINTENANCE_MESSAGE = env.str(
        "HELIPORT_CORE_MAINTENANCE_MESSAGE", default="No maintenance today"
    )

    #: Base URL for all links to "about" pages
    ABOUT_BASE_URL = "https://heliport.hzdr.de"
    #: Favicon URL used for public landing pages
    PUBLIC_FAVICON_URL = "https://heliport.hzdr.de/favicon.ico"

    #: Prefix used for generation of project identifiers
    HZDR_ID_PREFIX = "HZDR"

    #: The number of entries returned from autocomplete views
    AUTOCOMPLETE_COUNT = 10

    #: The name prefix for heliport groups
    GROUP_NAME_PREFIX = "heliportprojectgroup"  # -<group_id> will be appended

    #: The key used for encryption of credentials in the database.
    #: (The key shown in the docs is just the dummy value from the CI.)
    CRYPTO_KEY = env.str("HELIPORT_CORE_CRYPTO_KEY")
    if CRYPTO_KEY == "":
        raise ImproperlyConfigured("HELIPORT_CORE_CRYPTO_KEY must not be empty")

    class Meta:
        prefix = "HELIPORT_CORE"


class HeliportHandlesAppConf(AppConf):
    """Handle-related settings of the :mod:`heliport.core` app.

    All settings in this class can be overwritten in settings.py
    """

    REGISTER_HANDLES = env.bool("HELIPORT_HANDLE_REGISTER_HANDLES", default=False)
    USE_DUMMY_STRING_FOR_UNREGISTERED_HANDLES = settings.DEVELOPMENT
    DIGITAL_OBJECT_HOST = env.str(
        "HELIPORT_HANDLE_DIGITAL_OBJECT_HOST", default="https://vlsdms.fz-rossendorf.de"
    )
    LOCAL_DIGITAL_OBJECT_HOST = env.str(
        "HELIPORT_HANDLE_LOCAL_DIGITAL_OBJECT_HOST",
        default="https://vlsdms.fz-rossendorf.de",
    )
    URL_FORMAT = env.str("HELIPORT_HANDLE_URL_FORMAT", default=None)
    REGISTER_JSON_URL_HANDLES = env.bool(
        "HELIPORT_HANDLE_REGISTER_JSON_URL_HANDLES", default=True
    )
    RESOLVE_PID_FORMAT_TO = env.str(
        "HELIPORT_HANDLE_RESOLVE_PID_FORMAT_TO", default=None
    )
    PREFIX = env.str("HELIPORT_HANDLE_PREFIX", default="20.500.12865")
    SERVER_URL = env.str("HELIPORT_HANDLE_SERVER_URL", "https://handle.hzdr.de")
    API_URL = f"{SERVER_URL}/api/handles"
    CERTIFICATE_PATH = env.str(
        "HELIPORT_HANDLE_CERTIFICATE_PATH",
        default="/etc/ssl/certs/heliport_handle_admin.crt",
    )
    PRIVATE_KEY_PATH = env.str(
        "HELIPORT_HANDLE_PRIVATE_KEY_PATH",
        default="/etc/ssl/private/heliport_handle_admin.key",
    )
    DESCRIPTION = (
        "This handle is generated by HELIPORT at Helmholtz-Zentrum Dresden-Rossendorf. "
        "HELIPORT is a system in its early prototyping phase and not yet available to the public. "
        "If you are at HZDR and you have access rights, you can follow the URL below. "
    )

    class Meta:
        prefix = "HELIPORT_HANDLE"
