# SPDX-FileCopyrightText: 2023 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""
Register your actions in apps.py like this::

    from django.apps import AppConfig
    from heliport.core.digital_object_actions import actions

    class MyConfig(AppConfig):
        name = "my_app_name"

        def ready(self):
            from .views import MyAction, my_action_generation_func
            actions.register(MyAction)
            actions.register(my_action_generation_func)

where ``my_cation_generation_func`` is a function returning a list of actions, given
optionally a :class:`heliport.core.digital_object_interface.GeneralDigitalObject`, a
:class:`heliport.core.models.Project` and/or a
:class:`heliport.core.models.HeliportUser`, and `MyAction` is a subclass of
:class:`Action`.
"""


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, List, Type, Union

from heliport.core.utils.collections import DecoratorList

if TYPE_CHECKING:
    from heliport.core.digital_object_interface import GeneralDigitalObject
    from heliport.core.models import DigitalObject, HeliportUser, Project


class Action(ABC):
    """Represent an action that can be executed on some digital objects"""

    def __init__(
        self, obj: ObjType = None, project: Project = None, user: HeliportUser = None
    ):
        self.obj = obj
        self.project = project
        self.user = user

    @property
    @abstractmethod
    def name(self) -> str:
        """display name of this action"""

    @property
    @abstractmethod
    def icon(self) -> str:
        """name of a fontawesome icon"""

    @property
    @abstractmethod
    def link(self) -> str:
        """link to where the action can be executed"""

    @property
    @abstractmethod
    def applicable(self) -> bool:
        """determine if the action is executable for the given obj, project and user"""


if TYPE_CHECKING:
    ObjType = Union[DigitalObject, GeneralDigitalObject]
    ActionBuilder = Callable[[ObjType, Project, HeliportUser], List[Action]]


actions: DecoratorList[Union[Type[Action], ActionBuilder]] = DecoratorList()


def get_actions(
    obj: ObjType = None, project: Project = None, user: HeliportUser = None
) -> List[Action]:
    """Get all registered actions that are applicable for the given parameters."""
    result = []
    for action in actions:
        possible_actions = action(obj, project, user)
        if isinstance(possible_actions, Action):
            possible_actions = [possible_actions]

        result.extend(a for a in possible_actions if a.applicable)

    return result
