# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.urls import reverse

from heliport.core.app_interaction import Module


class ProjectConfigurationModule(Module):
    name = "Project Configuration"
    module_id = "project_configuration"
    icon = "fa-solid fa-gear"

    def get_url(self, project):
        return reverse("core:project", kwargs={"project": project.pk})

    def is_configured(self, project):
        return project.label


class Images(Module):
    name = "Images"
    module_id = "images"

    def get_url(self, project):
        return reverse("core:image_list", kwargs={"project": project.pk})

    def is_configured(self, project):
        from .models import Image

        return bool(Image.objects.filter(projects=project))


class Timeline(Module):
    name = "Project Timeline"
    module_id = "timeline"
    icon = "fa-solid fa-clock"

    def get_url(self, project):
        return reverse("core:project_timeline", kwargs={"project": project.pk})

    def is_configured(self, project):
        return True


class Tags(Module):
    name = "Tags"
    module_id = "tags"
    icon = "fa-solid fa-tags"

    def get_url(self, project):
        return reverse("core:tag_list", kwargs={"project": project.pk})

    def is_configured(self, project):
        from .models import Tag

        return bool(Tag.objects.filter(projects=project))


class SubProjects(Module):
    name = "Subprojects"
    module_id = "subprojects"
    icon = "fa-solid fa-table-cells-large"

    def get_url(self, project):
        return reverse("core:subprojects", kwargs={"project": project.pk})

    def is_configured(self, project):
        from .models import Project

        return bool(Project.objects.filter(projects=project))


def get_search_url():
    return reverse("core:search")


def serialize_project(project):
    def user_dict(user):
        if user is None:
            return None
        return {
            "datacite:hasIdentifier": user.authentication_backend_id,
            "datacite:orcid": user.orcid,
            "rdfs:label": user.display_name,
        }

    result = {
        "heliport:project_id": project.project_id,
        "datacite:hasIdentifier": project.hzdr_id,
        "heliport:uuid": project.uuid,
        f"datacite:{project.identifier_scheme.datacite_id}": project.identifier_link,
        "heliport:label": project.label,
        "time:hasBeginning": str(project.created),
        "datacite:hasDescription": project.description,
        "heliport:owner": user_dict(project.owner),
    }

    if project.group is not None:
        result["heliport:group"] = str(project.group)

    if project.deleted:
        result["time:hasEnd"] = str(project.deleted)

    co_owners = project.co_owners.all()
    if co_owners:
        result["heliport:co_owners"] = [user_dict(guy) for guy in co_owners]

    contributions = project.contribution_set.all()
    if contributions:
        result["datacite:contributors"] = [
            {
                k: v
                for k, v in {
                    "datacite:contributorType": contribution.type,
                    "datacite:name": contribution.contributor.display_name,
                    "datacite:nameType": "Personal",
                    "datacite:nameIdentifiers": [
                        item
                        for item in [
                            {
                                "datacite:nameIdentifier": contribution.contributor.orcid,
                                "datacite:nameIdentifierScheme": "ORCID",
                                "datacite:schemeURI": "http://orcid.org",
                            }
                            if contribution.contributor.orcid
                            else {},
                        ]
                        if item
                    ],
                    "datacite:affiliations": [
                        {"datacite:affiliation": contribution.contributor.affiliation}
                    ]
                    if contribution.contributor.affiliation
                    else [],
                }.items()
                if v
            }
            for contribution in contributions
        ]

    return result
