# SPDX-FileCopyrightText: 2023 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from django.apps import apps
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.shortcuts import get_object_or_404

from heliport.core import permissions
from heliport.core.app_interaction import (
    Module,
    generate_project_dropdown,
    generate_project_navbar,
)
from heliport.core.models import Project
from heliport.core.utils.context import Context, project_header_breadcrumbs

logger = logging.getLogger(__name__)


class HeliportLoginRequiredMixin(LoginRequiredMixin):
    login_url = "/login"
    permission_denied_message = "Please log in to view this site"
    raise_exception = False


class HeliportBreadcrumbsMixin:
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = Project.objects.filter(pk=self.kwargs["project"]).first()

        if project is None:
            return context

        reset = False
        breadcrumb_label = ""

        # Manually implemented method has the highest priority
        if hasattr(self, "get_breadcrumb_label_and_reset"):
            breadcrumb_label, reset = self.get_breadcrumb_label_and_reset()
        # This is the case if the current view is the entry point to an app
        elif (
            app_name := self.get_app_label(
                self.request.resolver_match.app_name, self.request.path, project
            )
        ) is not None:
            reset = True
            breadcrumb_label = app_name
        # This is the case in views that inherit from the generic DigitalObjectListView
        elif hasattr(self, "get_digital_object"):
            # This doesn't have the cache effect it is supposed to have. When
            # refactoring the mixins, make this Context a class member
            with Context(
                getattr(self.request.user, "heliportuser", None), project
            ) as cache:
                obj = self.get_digital_object(cache)
                if hasattr(obj, "label") and obj.label:
                    breadcrumb_label = obj.label
                elif hasattr(obj, "description") and obj.description:
                    breadcrumb_label = obj.description
        # This is the case in generic views deriving from Django's SingleObjectMixin
        elif hasattr(self, "get_object") and (obj := self.get_object()) != project:
            if hasattr(obj, "label") and obj.label:
                breadcrumb_label = obj.label
            elif hasattr(obj, "description") and obj.description:
                breadcrumb_label = obj.description

        context["project_header_breadcrumbs"] = project_header_breadcrumbs(
            self.request, breadcrumb_label, self.request.path, reset=reset
        )

        return context

    @staticmethod
    def get_app_label(app_name, path, project):
        app = apps.get_app_config(app_name)

        if not hasattr(app.module, "interface"):
            return None

        interface_module = app.module.interface
        # Instantiate all app interaction classes within the module, except from the
        # imported base class
        interfaces = [
            cls()
            for cls in interface_module.__dict__.values()
            if isinstance(cls, type) and issubclass(cls, Module) and cls != Module
        ]
        # Find the interface that matches the current URL
        interfaces = list(filter(lambda i: i.get_url(project) == path, interfaces))
        if interfaces:
            return interfaces[0].name

        return None


class HeliportProjectMixin(
    HeliportLoginRequiredMixin, UserPassesTestMixin, HeliportBreadcrumbsMixin
):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        context["project"] = project
        context["project_dropdown_modules"] = generate_project_dropdown(project)
        context["project_navbar_modules"] = generate_project_navbar(project)

        return context

    def test_func(self):
        user = self.request.user.heliportuser
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        return user in project.members()


class HeliportObjectMixin(
    HeliportLoginRequiredMixin, UserPassesTestMixin, HeliportBreadcrumbsMixin
):
    def test_func(self):
        user = self.request.user.heliportuser
        return permissions.has_permission_for(user, self.get_object())

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = get_object_or_404(Project, pk=self.kwargs["project"])

        context["project"] = project
        context["project_dropdown_modules"] = generate_project_dropdown(project)
        context["project_navbar_modules"] = generate_project_navbar(project)

        obj = self.get_object()
        if project not in obj.projects.all():
            logger.error(
                f"""incorrect project when opening {obj}.
            Project: {project.pk}; Available: {[p.pk for p in obj.projects.all()]}"""
            )
        return context
