# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import re
import typing
from base64 import b64decode
from collections import namedtuple
from mimetypes import guess_type
from pathlib import Path
from typing import Dict, Optional

from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q
from django.shortcuts import reverse
from django.template import Context as TemplateContext
from django.template import Template
from knox import models as knox_models
from rdflib import BNode, URIRef

import heliport.core.utils.handles as handles
from heliport.core.utils.collections import RootedRDFGraph
from heliport.core.utils.colors import pick_foreground_color
from heliport.core.utils.command_line_execution import create_command_executor
from heliport.core.utils.IDs import (
    find_identifier_scheme,
    generate_object_id,
    remove_special_chars,
    test_resolvability,
)
from heliport.core.utils.string_tools import remove_prefix
from heliport.core.utils.url import url_encode

from .digital_object_interface import GeneralDigitalObject
from .utils import encryption
from .utils.context import Context

logger = logging.getLogger(__name__)


class Group(models.Model):
    group_id = models.AutoField(primary_key=True)
    backend_id = models.CharField(max_length=100, null=True)  # noqa: DJ001
    display_name = models.CharField(max_length=100, blank=True)

    def __str__(self):
        if self.display_name:
            assert isinstance(self.display_name, str), f"invalid: {self.display_name}"
            return self.display_name
        if self.backend_id:
            return f"<Group {self.backend_id}>"
        return f"<Group {self.group_id}>"


class HeliportGroup(models.Model):
    group_id = models.AutoField(primary_key=True)
    group_name = models.TextField(default="UnnamedGroup")

    def __str__(self):
        return f"{self.group_name} ({self.group_id})"


class HeliportUser(models.Model):
    user_id = models.AutoField(primary_key=True)
    authentication_backend_id = models.CharField(  # noqa: DJ001
        max_length=100, null=True
    )
    auth_user = models.OneToOneField(
        User, on_delete=models.SET_NULL, null=True, blank=True
    )
    display_name = models.CharField(max_length=100, default="<User>")
    affiliation = models.CharField(max_length=500, blank=True)
    orcid = models.CharField(max_length=100, null=True, blank=True)  # noqa: DJ001
    email_is_public = models.BooleanField(default=False)
    email_sha1_is_public = models.BooleanField(default=True)
    uuid = models.UUIDField(
        primary_key=False, default=generate_object_id, editable=False
    )

    def __str__(self):
        return str(self.display_name)

    @property
    def email(self):
        if not self.email_is_public:
            return None
        if not self.auth_user:
            return None
        return self.auth_user.email

    def to_rdflib_node(self):
        if self.orcid:
            return URIRef(url_encode(self.orcid))
        else:
            return BNode(remove_special_chars(self.uuid))

    def set_orcid(self, orcid):
        orcid_link = "https://orcid.org/"
        if orcid and not orcid.startswith(orcid_link):
            orcid = f"{orcid_link}{orcid}"
        if orcid:
            self.orcid = orcid
        else:
            self.orcid = None


class DigitalObject(models.Model, GeneralDigitalObject):
    class ObjectPermissions(models.IntegerChoices):
        # Everybody can see and edit
        PUBLIC = 1
        # Private + Members of a project that this object is in can see and edit
        PROJECT = 2
        # only members of this Digital Object can see and edit
        PRIVATE = 3

    digital_object_id = models.AutoField(primary_key=True)
    persistent_id = models.CharField(  # noqa: DJ001
        max_length=500, null=True, blank=True
    )
    generated_persistent_id = models.CharField(  # noqa: DJ001
        max_length=500, null=True, blank=True
    )
    category = models.TextField()
    label = models.TextField()
    label_is_public = models.BooleanField(default=False)
    description = models.TextField(blank=True, default="")
    description_is_public = models.BooleanField(default=False)
    created = models.DateTimeField("date created", auto_now_add=True)
    last_modified = models.DateTimeField("date last modified", auto_now=True)
    last_modified_is_public = models.BooleanField(default=False)
    deleted = models.DateTimeField("date deleted", null=True, blank=True)
    permission = models.IntegerField(
        choices=ObjectPermissions.choices, default=ObjectPermissions.PROJECT
    )
    special_heliport_role = models.CharField(
        max_length=60, blank=True, null=True, unique=True
    )
    is_helper = models.BooleanField(default=False)

    projects = models.ManyToManyField("Project", blank=True, related_name="parts")
    projects_is_public = models.BooleanField(default=False)
    co_owners = models.ManyToManyField(
        HeliportUser, related_name="co_owned_digital_objects", blank=True
    )
    owner = models.ForeignKey(HeliportUser, on_delete=models.SET_NULL, null=True)
    members_is_public = models.BooleanField(default=False)
    heliport_groups = models.ManyToManyField(HeliportGroup, blank=True)

    def __str__(self):
        id_str = (
            self.pk
            if self.generated_persistent_id is None
            else self.generated_persistent_id
        )
        return f"{self.label} ({id_str})"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._context = None

    @staticmethod
    def type_id() -> str:
        """for resolution"""
        return "DigitalObject"

    def get_identifying_params(self) -> Dict[str, str]:
        """for resolution"""
        return {"digital_object_id": str(self.digital_object_id)}

    @staticmethod
    def resolve(
        params: Dict[str, str], context: Context
    ) -> Optional[GeneralDigitalObject]:
        """for resolution"""
        obj = DigitalObject.objects.filter(
            digital_object_id=params.get("digital_object_id")
        ).first()
        if obj is not None:
            obj = obj.casted_or_self()
        return obj

    def as_digital_object(self, context: Context):
        """For serialization as a
        :class:`heliport.core.digital_object_interface.GeneralDigitalObject`.
        """
        if self.digital_object_id is None:
            self.save()
        correctly_casted = self.casted_or_self()
        correctly_casted.set_context(context)
        return correctly_casted

    def as_digital_object_if_exists(self, context):
        """For serialization as a
        :class:`heliport.core.digital_object_resolution.Resolvable`.
        """
        return self

    def as_text(self) -> str:
        """For serialization as a
        :class:`heliport.core.utils.serialization.GeneralValue`.
        """
        if self.label:
            return self.label
        if self.description:
            return self.description
        if self.persistent_id:
            return self.persistent_id
        return "digital object"

    def as_rdf(self) -> RootedRDFGraph:
        """For serialization as a
        :class:`heliport.core.utils.serialization.GeneralValue`.
        """
        # Maybe also use :meth:`full_url` like in :meth:`to_rdflib_node`?
        if self.persistent_id:
            return RootedRDFGraph.from_uri_str(self.persistent_id)
        return RootedRDFGraph.from_blank()

    def as_html(self) -> str:
        """For serialization as a
        :class:`heliport.core.utils.serialization.GeneralValue`.
        """
        t = Template('<a href="{{ obj.identifier_link }}">{{ obj.as_text }}</a>')
        c = TemplateContext({"obj": self})
        return t.render(c)

    @property
    def viewing_user(self) -> typing.Optional[HeliportUser]:
        """The user that currently deals with this digital object.

        Information may be restricted to what the user can see. None means to only
        include public infos.
        """
        if self.context is not None:
            return self.context.user_or_none

    def set_context(self, context: Context):
        """
        Set an :class:`heliport.core.utils.context.Context` to be used by operations
        of this ``DigitalObject``. By setting the context you guarantee that,
        and you are responsible for, closing the context after all operations are done.
        It is not allowed to change the user or downgrade the permission back to public
        again. Create a new digital object instance for that.
        """
        assert self.viewing_user is None or self.viewing_user == context.user_or_none, (
            "It is not save to change the user like this, "
            "because some code might not recheck permissions. "
            "Please create a new instance for each user including the public user None."
        )
        self._context = context

    @property
    def context(self) -> typing.Optional[Context]:
        """
        If this function does not return ``None`` the return value can be used for
        optimizations where expensive setup and teardown steps can be shared between
        operations.
        See :class:`heliport.utils.collections.ExitContext` for more details.
        """
        return self._context

    def access(self, context: Context):
        """For serialization as a
        :class:`heliport.core.digital_object_interface.GeneralDigitalObject`.
        """
        from .permissions import PermissionChecker

        return PermissionChecker(self, context)

    @property
    def category_list(self):
        pk_list = self.category.split("/")
        return [
            DigitalObject.objects.filter(pk=pk).first() if pk.isdecimal() else None
            for pk in pk_list
        ]

    @category_list.setter
    def category_list(self, values):
        self.category = "/".join(str(c.pk) for c in values)

    @property
    def last_category(self):
        cl = self.category_list
        if not cl or not cl[-1]:
            return None
        else:
            return cl[-1]

    @property
    def category_str(self):
        return "/".join(
            c.label if c is not None else "None" for c in self.category_list
        )

    @category_str.setter
    def category_str(self, value):
        vocab = Vocabulary()

        namespace_list = value.split("/")
        namespace_category_str = f"HELIPORT/{vocab.Namespace.label}"
        last_namespace = None
        category_list = []
        for i, namespace_label in enumerate(namespace_list):
            if last_namespace is None:
                this_namespace = get_or_create_with_relation(
                    namespace_label,
                    namespace_category_str,
                    vocab.type,
                    vocab.Namespace,
                )
            else:
                this_namespace = get_or_create_with_relation(
                    namespace_label,
                    namespace_category_str,
                    vocab.subnamespace_of,
                    last_namespace,
                )
            assert_relation(this_namespace, vocab.type, vocab.Namespace)

            is_type = i == len(namespace_list) - 1
            if is_type:
                assert_relation(this_namespace, vocab.type, vocab.Class)

            category_list.append(this_namespace)
            last_namespace = this_namespace
        self.category_list = category_list

    @property
    def namespace_str(self):
        """a more appropriately named alias for :attr:`category_str`"""
        return self.category_str

    @namespace_str.setter
    def namespace_str(self, value: str):
        """setter"""
        self.category_str = value

    @property
    def category_in_persistent_id(self):
        pattern = ".*".join(c.label.lower() for c in self.category_list)
        return bool(re.search(pattern, str(self.persistent_id).lower()))

    def test_category_migrated(self):
        pk_list = self.category.split("/")
        return all(str(pk).isdecimal() for pk in pk_list)

    @property
    def relative_url(self):
        return reverse("core:landing_page", kwargs={"pk": self.digital_object_id})

    def local_url(self, request):
        return f"{request.scheme}://{request.get_host()}{self.relative_url}"

    def full_url(self):
        if settings.HELIPORT_HANDLE_URL_FORMAT:
            return settings.HELIPORT_HANDLE_URL_FORMAT.format(
                pk=self.digital_object_id, relative_url=self.relative_url
            )
        else:
            return (
                f"{ settings.HELIPORT_HANDLE_DIGITAL_OBJECT_HOST }{ self.relative_url }"
            )

    @property
    def identifier_link(self):
        if self.persistent_id and "://" in self.persistent_id:
            return self.persistent_id
        return self.full_url()

    @property
    def identifier_scheme(self):
        result = find_identifier_scheme(self.persistent_id)
        return result

    @property
    def identifier_label(self):
        return self.identifier_scheme.label

    @property
    def identifier_display(self):
        prefix = self.identifier_scheme.prefix
        result = self.identifier_link
        if result.startswith(prefix):
            result = result[len(prefix) :]
        return result

    def set_identifier(self, new_identifier):
        if new_identifier is None:
            self.persistent_id = None
            return

        if new_identifier.startswith("https://"):
            self.persistent_id = new_identifier
            return

        if test_resolvability(f"https://api.datacite.org/dois/{new_identifier}"):
            self.persistent_id = f"https://doi.org/{new_identifier}"
            return

        for variation in [
            new_identifier,
            f"https://hdl.handle.net/{new_identifier}",
            f"https://hdl.handle.net/{settings.HELIPORT_HANDLE_PREFIX}/{new_identifier}",
            f"https://n2t.net/{new_identifier}",
            f"https://{new_identifier}",
        ]:
            if test_resolvability(variation):
                self.persistent_id = variation
                return

        self.persistent_id = new_identifier

    def register_handle_if_applicable(self, request):
        if self.generated_persistent_id is None and not self.persistent_id:
            self.register_handle(request)

    def register_handle(self, request):
        assert (
            self.category
        ), "category is part of the handle and has to be set before registering the handle"
        assert (
            self.pk is not None
        ), "the primary key is part of the handle -> save the object before and after"
        assert (
            self.generated_persistent_id is None
        ), "this object already has a generated_persistent_id"

        category_list = self.category_str.split("/")
        handle_name = f'{".".join(category_list)}.{self.pk}'
        if not self.persistent_id:
            if settings.HELIPORT_HANDLE_REGISTER_HANDLES:
                if settings.HELIPORT_HANDLE_REGISTER_JSON_URL_HANDLES:
                    response = handles.register_json_handle(
                        handle_name,
                        self.category_str,
                        self.full_url(),
                        self.local_url(request),
                    )
                else:
                    response = handles.register_handle(handle_name, self.full_url())
                self.persistent_id = handles.get_handle_from_response(response)
                self.generated_persistent_id = self.persistent_id
            elif settings.HELIPORT_HANDLE_USE_DUMMY_STRING_FOR_UNREGISTERED_HANDLES:
                self.persistent_id = handle_name

        self.save()

    def members(self):
        result = set(self.co_owners.all())
        if self.owner:
            result.add(self.owner)
        return result

    @property
    def contributions(self):
        return self.contribution_set.all()

    @contributions.setter
    def contributions(self, contribution_values):
        if contribution_values is None:
            contribution_values = []
        current = set(self.contributions)
        target = set(contribution_values)
        target.discard(None)
        for old in current - target:
            old.delete()
        for new in target - current:
            new.contribution_to = self
            new.save()

    @staticmethod
    def not_subclass_query():
        return {
            f"{s.__name__.lower()}__isnull": True
            for s in DigitalObject.__subclasses__()
        }

    @property
    def is_subclass(self):
        return any(
            hasattr(self, s.__name__.lower()) for s in DigitalObject.__subclasses__()
        )

    def casted_as_subclass(self):
        # this works only for direct subclasses
        # because __subclasses__() lists only direct subclasses
        for s in DigitalObject.__subclasses__():
            if hasattr(self, s.__name__.lower()):
                return getattr(self, s.__name__.lower())
        return None

    def casted_or_self(self):
        potential_subclass = self.casted_as_subclass()
        if potential_subclass is not None:
            return potential_subclass
        else:
            return self

    def to_rdflib_node(self):
        """May be deprecated in the future in favour of :meth:`as_rdf`"""
        if self.persistent_id:
            return URIRef(url_encode(self.persistent_id))
        else:
            return URIRef(url_encode(self.full_url()))

    def save_with_handle(self, category, project, request):
        self.category_str = category
        self.save()
        self.projects.add(project)
        self.register_handle_if_applicable(request)

    @property
    def hdl_net_landing_page(self):
        hdl_landing_page = None
        if str(self.persistent_id).startswith("https://hdl.handle.net/"):
            hdl_landing_page = f"{self.persistent_id}?noredirect"

            local_handle_prefix = (
                f"https://hdl.handle.net/{settings.HELIPORT_HANDLE_PREFIX}/"
            )
            if hdl_landing_page.startswith(local_handle_prefix):
                suffix = remove_prefix(hdl_landing_page, local_handle_prefix)
                hdl_landing_page = f"{settings.HELIPORT_HANDLE_SERVER_URL}/{settings.HELIPORT_HANDLE_PREFIX}/{suffix}"
        return hdl_landing_page

    def create_copy(self):
        """Related Django docs: :ref:`django:topics/db/queries:copying model instances`"""
        obj = DigitalObject.objects.get(digital_object_id=self.digital_object_id)
        obj = obj.casted_or_self()
        obj._state.adding = True
        obj.pk = None
        obj.digital_object_id = None
        obj.generated_persistent_id = None
        obj.persistent_id = None
        if hasattr(obj, "digitalobject_ptr"):
            base_obj = DigitalObject.objects.get(
                digital_object_id=obj.digitalobject_ptr_id
            )
            base_obj._state.adding = True
            base_obj.pk = None
            base_obj.save()
            obj.digitalobject_ptr = base_obj
        obj.save()
        if obj.pk is None:
            if obj.digital_object_id:
                logger.warning(
                    f"Django somehow failed to set pk of object {obj.digital_object_id} when saving the object"
                )
                obj = DigitalObject.objects.get(digital_object_id=obj.digital_object_id)
                obj = obj.casted_or_self()
            else:
                logger.error(
                    "Django somehow failed to set obj.pk and obj.digital_object_id"
                )

        for rel in self.relations.all():
            rel._state.adding = True
            rel.pk = None
            rel.subject = obj
            rel.save()
        for attr in self.attributes.all():
            attr._state.adding = True
            attr.pk = None
            attr.subject = obj
            attr.save()

        DigitalObjectRelation.objects.create(
            subject=obj, predicate=Vocabulary().is_version_of, object=self
        )
        return obj

    def add_tag(self, tag):
        """
        Tag the object with ``tag``. See also :meth:`Tag.set_on`.

        :type tag: Tag
        """
        tag.set_on(self)

    def remove_tag(self, tag):
        """
        Object is no longer tagged with ``tag``. See also :meth:`Tag.unset_on`.

        :type tag: Tag
        """
        tag.unset_on(self)

    def get_tags(self, project=None):
        from .utils.queries import tags_of

        return tags_of(self, project)


class DigitalObjectRelation(models.Model):
    digital_object_relation_id = models.AutoField(primary_key=True)
    subject = models.ForeignKey(
        DigitalObject, on_delete=models.CASCADE, related_name="relations"
    )
    predicate = models.ForeignKey(
        DigitalObject, on_delete=models.CASCADE, related_name="predicate_relations"
    )
    object = models.ForeignKey(
        DigitalObject, on_delete=models.CASCADE, related_name="reverse_relations"
    )
    is_public = models.BooleanField(default=False)
    special_heliport_role = models.CharField(  # noqa: DJ001
        max_length=15, blank=True, null=True
    )

    def __str__(self):
        return f"{self.pk}: {self.subject.label} --[{self.predicate.label}]--> {self.object.label}"


class DigitalObjectAttributes(models.Model):
    digital_object_attribute_id = models.AutoField(primary_key=True)
    subject = models.ForeignKey(
        DigitalObject, on_delete=models.CASCADE, related_name="attributes"
    )
    predicate = models.ForeignKey(
        DigitalObject, on_delete=models.CASCADE, related_name="predicate_attributes"
    )
    value = models.TextField()
    is_public = models.BooleanField(default=False)
    special_heliport_role = models.CharField(  # noqa: DJ001
        max_length=15, blank=True, null=True
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["subject", "predicate", "special_heliport_role"],
                condition=models.Q(special_heliport_role__isnull=False),
                name="unique_special_attributes_per_subject",
            )
        ]

    def __str__(self):
        return self.value


class Project(DigitalObject):
    project_id = models.AutoField(primary_key=True)
    hzdr_id = models.CharField(max_length=100)
    uuid = models.UUIDField(
        primary_key=False, default=generate_object_id, editable=False
    )
    group = models.ForeignKey(Group, on_delete=models.SET_NULL, null=True)
    parts_is_public = models.BooleanField(default=False)

    original_id = models.CharField(max_length=100, null=True, blank=True)  # noqa: DJ001

    def get_custom_view_url(self, user=None):
        """This defines the custom landing page to be the object graph"""
        if user in self.members():
            return reverse("core:project_graph", kwargs={"project": self.pk})
        return None

    def digital_object_lookup(self):
        """get parts as dict of query sets keyed by type as string"""
        result = {}
        for s in DigitalObject.__subclasses__():
            name = s.__name__.lower()
            query = {f"{name}__isnull": False}

            result[name] = self.parts.filter(**query)
        return result

    def last_modified_aggregate(self):
        """
        Get the last date this project or one of its parts was modified.
        Note that parts in subprojects are not counted.
        """
        if self.project_id is None:
            return None

        last_modified = self.parts.aggregate(models.Max("last_modified"))[
            "last_modified__max"
        ]
        if last_modified is None or self.last_modified is None:
            return self.last_modified
        else:
            return max(last_modified, self.last_modified)

    def sub_project_list(self, max_level=8, level=1, parent_path=None):
        """
        Get a list of named tuples with the following attributes:

            - path: list of parent project ids
            - level: how deep is the subproject nested (as iterable python range)
            - project: the subproject model instance
            - has_child: boolean indicating if there are more subprojects under this

        for each subproject
        """
        if level > max_level:
            return []
        if parent_path is None:
            parent_path = []

        SubprojectItem = namedtuple("SubprojectItem", "path level project has_child")

        result = []
        for sub_project in Project.objects.filter(projects=self, deleted__isnull=True):
            sub_result = sub_project.sub_project_list(
                max_level, level + 1, [*parent_path, f"_{sub_project.project_id}"]
            )
            result.append(
                SubprojectItem(parent_path, range(level), sub_project, bool(sub_result))
            )
            result.extend(sub_result)

        return result

    def get_parent_projects(self, exclude: typing.Iterable = None):
        """
        The set of all projects that have this project as subproject. Ignore subprojects
        and subprojects of subprojects that are in the ``exclude`` set.
        The project itself is also part of the result.
        """
        new_exclude = {self}
        if exclude is not None:
            new_exclude.update(exclude)

        result = {self}
        for parent_project in self.projects.all():
            if parent_project not in new_exclude:
                result.update(parent_project.get_parent_projects(new_exclude))

        return result


class LoginInfo(models.Model):
    class LoginTypes(models.TextChoices):
        SSH = "ssh connection"
        USER_PASSWORD = "username and password"
        TOKEN = "authentication token"

    login_info_id = models.AutoField(primary_key=True)
    user = models.ForeignKey(HeliportUser, on_delete=models.CASCADE)
    username = models.CharField(max_length=100, blank=True)
    encrypted_password = models.CharField(max_length=500, blank=True)
    key = models.TextField(blank=True)
    machine = models.CharField(max_length=100, blank=True)
    name = models.CharField(max_length=100)
    type = models.CharField(max_length=50, choices=LoginTypes.choices)
    data = models.TextField(blank=True)
    via = models.ForeignKey("self", on_delete=models.SET_NULL, null=True, blank=True)
    slurm = models.BooleanField(default=False)

    def __str__(self):
        return self.name

    @property
    def password(self):
        if self.encrypted_password:
            return encryption.decrypt(self.encrypted_password)
        else:
            return ""

    @password.setter
    def password(self, value):
        if value:
            self.encrypted_password = encryption.encrypt(value)
        else:
            self.encrypted_password = ""

    def build_command_executor(self, depth=10):
        if self.type != self.LoginTypes.SSH:
            raise TypeError("not a ssh login")
        if depth <= 0:
            raise ValueError("max depth reached: may be caused by cyclic ssh login")

        indirection = None
        if self.via is not None:
            indirection = self.via.build_command_executor(depth=depth - 1)

        return create_command_executor(
            host=self.machine,
            username=self.username,
            key=self.key,
            indirection=indirection,
            auto_connect=False,
        )

    def build_command_executor_indirection(self):
        if self.via is None:
            return None

        return self.via.build_command_executor()

    @property
    def connected(self) -> bool:
        if self.type != self.LoginTypes.SSH:
            return False
        else:
            return bool(self.data) and bool(self.key)

    def get_to_connect_list(self, depth=10, exclude_self=False) -> list:
        if self.type != self.LoginTypes.SSH:
            raise TypeError("not a ssh login")
        if depth <= 0:
            raise ValueError("max depth reached: may be caused by cyclic ssh login")

        result = []
        if self.via is not None:
            result = self.via.get_to_connect_list(depth=depth - 1)
        if not self.connected and not exclude_self:
            result.append(self)
        return result

    def get_unconnected_parents(self) -> list:
        return self.get_to_connect_list(exclude_self=True)


class Contribution(models.Model):
    class ContributionTypes(models.TextChoices):
        CONTACT_PERSON = "ContactPerson", "Contact Person"
        DATA_COLLECTOR = "DataCollector", "Data Collector"
        DATA_CURATOR = "DataCurator", "Data Curator"
        DATA_MANAGER = "DataManager", "Data Manager"
        DISTRIBUTOR = "Distributor", "Distributor"
        EDITOR = "Editor", "Editor"
        HOSTING_INSTITUTION = "HostingInstitution", "Hosting Institution"
        PRODUCER = "Producer", "Producer"
        PROJECT_LEADER = "ProjectLeader", "Project Leader"
        PROJECT_MANAGER = "ProjectManager", "Project Manager"
        PROJECT_MEMBER = "ProjectMember", "Project Member"
        REGISTRATION_AGENCY = "RegistrationAgency", "Registration Agency"
        REGISTRATION_AUTHORITY = "RegistrationAuthority", "Registration Authority"
        RELATED_PERSON = "RelatedPerson", "Related Person"
        RESEARCHER = "Researcher", "Researcher"
        RESEARCH_GROUP = "ResearchGroup", "Research Group"
        RIGHTS_HOLDER = "RightsHolder", "Rights Holder"
        SPONSOR = "Sponsor", "Sponsor"
        SUPERVISOR = "Supervisor", "Supervisor"
        WORK_PACKAGE_LEADER = "WorkPackageLeader", "Work Package Leader"
        OTHER = "Other", "Other"

    contribution_id = models.AutoField(primary_key=True)
    contributor = models.ForeignKey(HeliportUser, on_delete=models.CASCADE)
    contribution_to = models.ForeignKey(DigitalObject, on_delete=models.CASCADE)
    type = models.CharField(
        max_length=50,
        choices=ContributionTypes.choices,
        default=ContributionTypes.PROJECT_MEMBER,
    )
    uuid = models.UUIDField(
        primary_key=False, default=generate_object_id, editable=False
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["contributor", "contribution_to"],
                name="unique_digital_object_contribution",
            )
        ]

    def __str__(self):
        return f"{self.contributor}/{self.contribution_to}"

    def to_rdflib_node(self):
        return BNode(remove_special_chars(self.uuid))

    @classmethod
    def type_from_str(cls, the_type):
        for t, label in cls.ContributionTypes.choices:
            if t == the_type or label == the_type:
                return t
        return None


class Image(DigitalObject):
    class ImageStates(models.IntegerChoices):
        OK = 1
        UPLOADING = 2
        FAILED = 3

    directory = models.TextField()
    filename = models.TextField()
    login = models.ForeignKey(LoginInfo, on_delete=models.SET_NULL, null=True)
    status = models.IntegerField(
        choices=ImageStates.choices, default=ImageStates.UPLOADING
    )
    message = models.TextField(blank=True)
    replace_if_exists = models.BooleanField(default=False)  # always replacing currently
    download_url_is_public = models.BooleanField(default=False)

    @property
    def path(self):
        return Path(self.directory) / self.filename

    @property
    def mimetype(self):
        result, encoding = guess_type(self.filename)
        return result

    def fail(self, message):
        self.status = self.ImageStates.FAILED
        self.message = message
        self.save()
        return None

    def upload_image(self, image_url):
        self.status = self.ImageStates.UPLOADING
        self.save()

        data_url_parts = image_url.split(";base64,", maxsplit=1)
        if len(data_url_parts) != 2:
            logger.warning(f"Invalid base64 data url: {image_url[:40]}")
            return self.fail("Invalid base64 data url.")

        with self.login.build_command_executor() as executor:
            with executor.file(str(self.path), "wb") as file:
                file.write(b64decode(data_url_parts[1]))

        self.status = self.ImageStates.OK
        self.save()

    def download_data(self):
        try:
            with self.login.build_command_executor() as executor:
                with executor.file(str(self.path)) as file:
                    return file.read()
        except FileNotFoundError:
            if self.status == self.ImageStates.OK:
                self.status = self.ImageStates.FAILED
                self.message = "file not found"
                self.save()
            return b""

    def get_custom_url(self, user=None):
        if user in self.members():
            return reverse("core:image_data", kwargs={"pk": self.pk})
        return None


class Tag(DigitalObject):
    """
    A Digital Object D is tagged with a tag if relation (D, attribute, value) exists
    """

    tag_id = models.AutoField(primary_key=True)
    attribute = models.ForeignKey(
        DigitalObject,
        on_delete=models.CASCADE,
        related_name="tags_where_property",
        null=True,
    )
    value = models.ForeignKey(
        DigitalObject,
        on_delete=models.CASCADE,
        related_name="tags_where_value",
        null=True,
    )
    hex_color = models.CharField(max_length=8)
    requires_type = models.ForeignKey(
        DigitalObject,
        on_delete=models.CASCADE,
        null=True,
        related_name="tags_where_type_required",
    )
    last_use_date = models.DateTimeField(auto_now=True)

    @property
    def html_color(self):
        """color to be used in html e.g. #00FF35"""
        return f"#{self.hex_color}"

    @html_color.setter
    def html_color(self, value):
        """setter for :attr:`html_color`"""
        if value.startswith("#"):
            value = value[1:]
        self.hex_color = value

    @property
    def html_foreground_color(self):
        """color to be used for text if background has :attr:`html_color`"""
        return pick_foreground_color(self.html_color)

    @property
    def rgb_color(self):
        """same as :attr:`html_color` but returns a tuple of values up to 255"""
        try:
            return tuple(int(self.hex_color[i : i + 2], 16) for i in (0, 2, 4))
        except ValueError:
            return 0, 0, 0

    @rgb_color.setter
    def rgb_color(self, value):
        """setter for :attr:`rgb_color`"""
        r, g, b = value
        self.hex_color = f"{r:02x}{g:02x}{b:02x}"

    @property
    def requires_type_display(self):
        """
        Tag might be for specific type or any type. This property can be used in an
        HTML template
        """
        if self.requires_type is None:
            return "Anything"
        return self.requires_type

    def update_usage_date(self):
        """set last used to now"""
        self.save()

    def get_key_value(self):
        """
        Return tuple of attribute and value.
        Attribute defaults to has_tag and value to the tag itself if they are None.
        """
        vocab = Vocabulary()
        attribute = self.attribute
        value = self.value
        if attribute is None:
            attribute = vocab.has_tag
        if value is None:
            value = self
        return attribute, value

    def relation_query_set(self):
        attribute, value = self.get_key_value()
        vocab = Vocabulary()
        return DigitalObjectRelation.objects.filter(
            Q(predicate=attribute, object=value)
            | Q(predicate=vocab.has_tag, object=self)
        )

    def applicable_for(self, obj: DigitalObject):
        """determine if ``obj`` has type required by the tag"""
        if self.requires_type is None:
            return True
        vocab = Vocabulary()
        return DigitalObjectRelation.objects.filter(
            subject=obj, predicate=vocab.type, object=self.requires_type
        ).exists()

    def set_on(self, obj: DigitalObject):
        """Tag ``obj`` with this tag. See also :meth:`DigitalObject.add_tag`."""
        attribute, value = self.get_key_value()
        assert_relation(obj, attribute, value)
        self.update_usage_date()

    def unset_on(self, obj: DigitalObject):
        """Remove tag from ``obj``. See also :meth:`DigitalObject.remove_tag`."""
        vocab = Vocabulary()
        attribute, value = self.get_key_value()
        assert_not_relation(obj, attribute, value)
        assert_not_relation(obj, vocab.has_tag, self)
        self.update_usage_date()

    @property
    def as_dict(self):
        """Return a serialized version of this tag as dict to be converted to JSON."""
        return {
            "label": self.label,
            "tag_id": self.tag_id,
            "attribute": self.attribute_id,
            "value": self.value_id,
            "requires_type": self.requires_type_id,
            "color": self.html_color,
            "foreground": self.html_foreground_color,
        }


class NamedToken(models.Model):
    token_id = models.AutoField(primary_key=True)
    name = models.TextField()
    auth_token = models.OneToOneField(knox_models.AuthToken, on_delete=models.CASCADE)

    def __str__(self):
        return self.name


class MetadataField:
    def __init__(self, attribute_vocab, normalizer=str, *, role_name_in_db="attribute"):
        self.attribute = attribute_vocab
        self.normalizer = normalizer
        self.role = role_name_in_db

    def query(self, instance):
        return {
            "subject": instance,
            "predicate": self.attribute.obj,
            "special_heliport_role": self.role,
        }

    def __get__(self, instance, owner=None):
        attribute_object = DigitalObjectAttributes.objects.filter(
            **self.query(instance)
        ).first()
        if attribute_object is None:
            return None
        else:
            return attribute_object.value

    def __set__(self, instance, value):
        if value is None:
            attr = self.__get__(instance)
            if attr is not None:
                attr.delete()
        else:
            attr, is_new = DigitalObjectAttributes.objects.get_or_create(
                **self.query(instance)
            )
            attr.value = self.normalizer(value)
            attr.save()


class FromDB:
    """Get a digital object by special_heliport_role"""

    def __init__(self, role_name_in_db):
        self.role_name_in_db = role_name_in_db

    @property
    def obj(self):
        result = DigitalObject.objects.filter(
            special_heliport_role=self.role_name_in_db
        ).first()
        assert (
            result is not None
        ), f"add {self.role_name_in_db} to vocabulary_core.py and migrate db"
        return result

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return self.obj


class Vocabulary:
    """To get a digital object use `Vocabulary.<name>.obj` or `Vocabulary().<name>`"""

    Property = FromDB("Property")
    Class = FromDB("Class")
    Namespace = FromDB("Namespace")
    HELIPORT = FromDB("HELIPORT")
    Project = FromDB("Project")

    type = FromDB("type")
    suggested_property = FromDB("suggested_property")
    subnamespace_of = FromDB("subnamespace_of")
    subclass_of = FromDB("subclass_of")
    label = FromDB("label")
    description = FromDB("description")
    sub_property_of = FromDB("sub_property_of")
    has_tag = FromDB("has_tag")
    has_part = FromDB("has_part")
    input_of = FromDB("input_of")
    # input parameter in a workflow description
    has_input = FromDB("has_input")
    # something is used as the input parameter defined by has_input
    described_by_parameter = FromDB("described_by_parameter")
    # a job is described by some workflow/tool
    described_by_process = FromDB("described_by_process")
    primary_topic = FromDB("primary_topic")
    status = FromDB("status")
    has_identifier = FromDB("has_identifier")
    filename = FromDB("filename")
    is_version_of = FromDB("is_version_of")


def digital_object_form_label_and_pk(label, pk) -> typing.Optional[DigitalObject]:
    """
    Resolve a DigitalObject from pk.
    If pk is not truthy, label is used if it uniquely identifies a DigitalObject.
    This is useful for getting a digital object from an autocomplete.
    if pk is not an integer or a string representing an integer and label is not truthy,
    this function returns ``None``.
    """
    result = None
    if pk:
        result = DigitalObject.objects.filter(pk=pk).first()
    if result is None or label not in result.label:
        possibilities = []
        if label:
            possibilities = DigitalObject.objects.filter(label=label)
        if len(possibilities) == 1:
            result = list(possibilities)[0]

    return result


def get_or_create_with_relation(label, category, pred, obj):
    rel = DigitalObjectRelation.objects.filter(
        subject__label=label, predicate=pred, object=obj
    ).first()
    if rel is None:
        new_obj = DigitalObject.objects.create(label=label)
        rel, created = DigitalObjectRelation.objects.get_or_create(
            subject=new_obj, predicate=pred, object=obj
        )
        new_obj.category_str = category  # note the recursion: category_str calls get_or_create_with relation
        new_obj.save()
    return rel.subject


def assert_relation(subj: DigitalObject, pred: DigitalObject, obj: DigitalObject):
    """
    Insert the triple (subj, pred, obj) into :class:`DigitalObjectRelation` table.
    Semantically this makes the statement considered true in HELIPORT.
    Another way to view this is to set the key-value-pair "pred: obj" in the metadata
    of subj.

    Consider the example, where ``HELIPORT``, ``uses`` and ``Django``
    are :class:`DigitalObject` instances::

        assert_relation(HELIPORT, uses, Django)

    This means that the statement "HELIPORT uses Django" is assumed to be true or that
    the metadata of HELIPORT includes the key-value-pair "uses: Django".
    """
    rel, created = DigitalObjectRelation.objects.get_or_create(
        subject=subj, predicate=pred, object=obj
    )
    return created


def assert_not_relation(subj: DigitalObject, pred: DigitalObject, obj: DigitalObject):
    """
    Opposite of :func:`assert_relation`.
    Means that the statement is no longer assumed to be true.
    """
    try:
        rel = DigitalObjectRelation.objects.get(
            subject=subj, predicate=pred, object=obj
        )
        rel.delete()
        return True
    except DigitalObjectRelation.DoesNotExist:
        return False


def assert_relation_truth(
    subj: DigitalObject, pred: DigitalObject, obj: DigitalObject, truth: bool
):
    """
    same as :func:`assert_relation` and :func:`assert_not_relation` but which one is
    decided by ``truth``
    """
    if truth:
        return assert_relation(subj, pred, obj)
    else:
        return assert_not_relation(subj, pred, obj)


def get_special_object(special_heliport_role: str, allow_create_with: dict = None):
    if allow_create_with is None:
        return DigitalObject.objects.filter(
            special_heliport_role=special_heliport_role
        ).first()
    else:
        return DigitalObject.objects.get_or_create(
            defaults=allow_create_with, special_heliport_role=special_heliport_role
        )
