# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import hashlib
import json
import re
import typing
from abc import abstractmethod
from collections import defaultdict
from itertools import chain, zip_longest

from accept_types import get_best_match
from datacite import schema43
from django.conf import settings
from django.http import HttpResponse, JsonResponse
from django.shortcuts import reverse
from django.utils import timezone
from rdflib import DCTERMS, FOAF, RDF, SDO, BNode, Graph, Literal, URIRef
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework.fields import Field as RestField
from rest_framework.fields import empty as empty_rest_data
from rest_framework.fields import html as html_utils

from heliport.core.utils.collections import DecoratorDict
from heliport.core.utils.IDs import remove_special_chars

from .attribute_description import (
    Attribute,
    AttrObj,
    BaseAttribute,
    BoundAttribute,
    ContributionT,
    DateT,
    DigitalObjectT,
    DynamicAttribute,
    FunctionResultAttribute,
    InternalValueDescription,
    ListT,
    ManyAttribute,
    PersonT,
    ProjectT,
    StringT,
    TypeAttribute,
    UrlT,
)
from .models import (
    Contribution,
    DigitalObject,
    DigitalObjectAttributes,
    DigitalObjectRelation,
    Group,
    HeliportUser,
    Image,
    LoginInfo,
    Project,
    Vocabulary,
)
from .permissions import has_read_permission, has_write_permission
from .utils.json_ld import TargetDescription, reverse_context_mapping
from .vocabulary import FRAPO, FRBR, PRO, SCORO, SemanticDesktop
from .vocabulary import namespaces as custom_namespaces
from .vocabulary_descriptor import LiteralTriple, RelationTriple

#: collection of all metadata descriptions
static_attributes = DecoratorDict()
datacite_serializers = DecoratorDict()
rdf_serializers = DecoratorDict()
renderers = DecoratorDict()


def register_digital_object_attributes(cls):
    def decorator(func):
        def attributes():
            return digital_object_attributes() + func()

        static_attributes[cls] = attributes
        return attributes

    return decorator


class ProjectSerializer(serializers.ModelSerializer):
    class Meta:
        model = Project
        fields = [
            "project_id",
            "label",
            "created",
            "deleted",
            "description",
            "group",
            "owner",
            "co_owners",
            "persistent_id",
        ]


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = HeliportUser
        fields = [
            "user_id",
            "authentication_backend_id",
            "display_name",
            "orcid",
            "affiliation",
        ]


class GroupSerializer(serializers.ModelSerializer):
    class Meta:
        model = Group
        fields = ["display_name"]


class ContributionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Contribution
        fields = ["contribution_id", "type", "contributor", "contribution_to"]


class TokenSerializer(serializers.ModelSerializer):
    token = serializers.CharField(source="key")

    class Meta:
        model = LoginInfo
        fields = ["login_info_id", "name", "token"]


class DigitalObjectListSerializer(serializers.ListSerializer):
    def to_representation(self, digital_objects):
        result = []
        for digital_object in digital_objects:
            digital_object = digital_object.casted_or_self()
            serializer = DigitalObjectAPISerializer(
                digital_object, context=self.context
            )
            serializer.bind("", self)
            result.append(serializer.to_representation(digital_object))
        return result

    def update(self, instance, validated_data):
        raise NotImplementedError("update digital object list")

    def create(self, validated_data):
        raise NotImplementedError("create digital object list")


class DigitalObjectAPISerializer(serializers.ModelSerializer):
    class ID_KEY:
        in_json = "@id"
        in_obj = "identifier_link"

    @property
    def user(self):
        auth_user = None
        request = self.context.get("request")
        if request and hasattr(request, "user"):
            auth_user = request.user
        result = None
        if hasattr(auth_user, "heliportuser"):
            result = auth_user.heliportuser
        return result

    def validate(self, data):
        if self.instance is not None and not has_write_permission(
            self.user, self.instance
        ):
            raise ValidationError("You don't have write permission for this object.")
        return super().validate(data)

    def __init__(self, instance=None, data=empty_rest_data, **kwargs):
        if isinstance(instance, DigitalObject):
            instance = instance.casted_or_self()
        super().__init__(instance, data, **kwargs)
        self.fields[self.ID_KEY.in_json] = serializers.CharField(
            source=self.ID_KEY.in_obj, required=False
        )

        self.view = self.context.get("view")
        context = self.view.ld_context

        if isinstance(instance, DigitalObject):
            attributes = get_attributes_as(self.user, instance)
        else:
            attributes = get_static_attributes(DigitalObject)

        attr_index = defaultdict(list)
        for attr in attributes:
            key = (attr.label, attr.persistent_id)
            attr_index[key].append(attr)

        key_mapping = {}
        if data is not empty_rest_data:
            key_mapping = dict(
                reverse_context_mapping(
                    [
                        TargetDescription(label, persistent_id, label)
                        for label, persistent_id in attr_index
                    ],
                    {key: key for key in data},
                )
            )

        for (label, persistent_id), attrs in attr_index.items():
            context.tentative_set(
                self.fields,
                key_mapping.get(label, label),
                MetadataRestField(label, persistent_id, attrs, self.user),
                persistent_id,
            )

        if data is not empty_rest_data:
            data_context = data.get("@context")
            if not isinstance(data_context, dict):
                data_context = {}
            for key in data:
                key = str(key)
                if (
                    key.startswith("https://")
                    or key.startswith("http://")
                    or key in data_context
                ):
                    context.tentative_set(
                        self.fields,
                        key,
                        MetadataRestField(
                            key, data_context.get(key, key), [], self.user
                        ),
                        data_context.get(key, key),
                    )
        self.fields["New Property"] = TemplateField("core/api/new_property.html")

    def set_values(self, instance, validated_data):
        assert has_write_permission(self.user, instance), "no write permission"
        for key, field in self.fields.items():
            if key == self.ID_KEY.in_json:
                instance.persistent_id = validated_data.get(self.ID_KEY.in_obj)
            elif isinstance(field, MetadataRestField):
                field.set_values(instance, validated_data.get(key))
        instance.save()

    def update(self, instance, validated_data):
        self.set_values(instance, validated_data)
        return instance

    def create(self, validated_data):
        assert isinstance(self.user, HeliportUser)
        new_do = DigitalObject()
        new_do.save()
        new_do.co_owners.add(self.user)
        self.set_values(new_do, validated_data)
        return new_do

    def __getitem__(self, key):
        field = self.fields[key]
        value = self.data.get(key)
        error = self.errors.get(key) if hasattr(self, "_errors") else None

        if isinstance(field, MetadataRestField):
            return MetadataBoundField(field, value, error, self.instance)
        else:
            return super().__getitem__(key)

    class Meta:
        model = DigitalObject
        list_serializer_class = DigitalObjectListSerializer
        fields = [
            "digital_object_id",
        ]


class MetadataBoundField(serializers.BoundField):
    def __init__(self, field, value, error, instance):
        super().__init__(field, value, error)
        self.instance = instance

    def as_form_field(self):
        value = "" if (self.value is None or self.value is False) else self.value
        return self.__class__(self._field, value, self.errors, self.instance)

    def render_form_inputs(self):
        for attr in self.attrs:
            yield from attr.to_inputs(self.field_name, self.instance)


class TemplateField(RestField):
    def __init__(self, template):
        super().__init__(required=False, allow_null=True, write_only=True)
        self.style["template"] = template

    def to_internal_value(self, data):
        return None

    def to_representation(self, value):
        return None

    def get_attribute(self, instance):
        return None


class MetadataRestField(RestField):
    def __init__(
        self,
        label,
        persistent_id,
        attrs: typing.List[BaseAttribute],
        user: HeliportUser,
    ):
        super().__init__(
            required=False,
            allow_null=True,
            read_only=attrs and all(attr.read_only for attr in attrs),
        )
        self.key = label
        self.label = label
        self.persistent_id = persistent_id
        self.attrs = attrs
        self.user = user
        self.style["template"] = "core/form_field/metadata_rest_field.html"
        if len(self.attrs) > 0:
            self.label = self.attrs[0].label
        predicates = [
            attr.attribute
            for attr in self.attrs
            if isinstance(attr.attribute, DigitalObject)
        ]
        self.predicate_obj = min(predicates, key=lambda p: p.pk, default=None)

    @property
    def source_attrs(self):
        return [self.field_name]

    @source_attrs.setter
    def source_attrs(self, value):
        pass  # source_attrs are set to self.field_name.split(".") in bind() -> ignore it

    def get_or_generate_predicate_obj(self):
        if self.predicate_obj is None:
            self.predicate_obj = (
                DigitalObject.objects.filter(persistent_id=self.persistent_id)
                .order_by("digital_object_id")
                .first()
            )
        if self.predicate_obj is None:
            self.predicate_obj = DigitalObject(
                persistent_id=self.persistent_id, label=self.label
            )
            self.predicate_obj.save()
        return self.predicate_obj

    @property
    def writable_attrs(self):
        return (a for a in self.attrs if not a.read_only)

    @property
    def read_only_attrs(self):
        return (a for a in self.attrs if a.read_only)

    @property
    def writable_list_attrs(self):
        return (a for a in self.writable_attrs if a.handles_list)

    @property
    def writable_single_attrs(self):
        return (a for a in self.writable_attrs if not a.handles_list)

    def split_value_list(self, value_list):
        i = len(list(self.writable_single_attrs))
        return value_list[:i], value_list[i:]

    def get_value(self, dictionary):
        if html_utils.is_html_input(dictionary):
            html_result = dictionary.getlist(self.field_name, [])
            for i, v in enumerate(html_result):
                if isinstance(v, str) and v.isdecimal():
                    html_result[i] = int(v)
            return html_result
        result = super().get_value(dictionary)
        if not isinstance(result, list):
            result = [result]
        return result

    def to_internal_value(self, data):
        result = []

        single_data, list_data = self.split_value_list(data)
        for attr, d in zip_longest(self.writable_single_attrs, single_data):
            if d is empty_rest_data:
                d = None
            result.append(attr.deserialize(d))

        try:
            list_attrs = iter(self.writable_list_attrs)
            deserialized = next(list_attrs).deserialize(
                [None if d is empty_rest_data else d for d in list_data]
            )
            if isinstance(deserialized, InternalValueDescription):
                result.append(deserialized)
            else:
                result.extend(deserialized)
        except StopIteration:
            for d in list_data:
                t = DigitalObjectT() if isinstance(d, (int, dict)) else StringT()
                result.append(t.deserialize(d, {}))

        return result

    def to_representation(self, value):
        result = []
        has_list_attribute = False
        for attr in chain(
            self.writable_single_attrs,
            self.writable_list_attrs,
            self.read_only_attrs,
        ):
            serialized = attr.to_json(value)
            if attr.handles_list:
                has_list_attribute = True
                if serialized is not None:
                    result.extend(serialized)
            else:
                result.append(serialized)

        if len(result) == 1 and not has_list_attribute:
            result = result[0]
        return result

    def get_attribute(self, instance):
        return instance

    def set_values(self, instance, validated_value_list):
        assert has_write_permission(self.user, instance), "no permission"
        if validated_value_list is None:
            validated_value_list = []
        value_list = []
        for value in validated_value_list:
            if not isinstance(value, InternalValueDescription):
                value_list.append(value)
            elif value.describes_list:
                value_list.extend(value.get_value(instance))
            else:
                value_list.append(value.get_value(instance))

        single_values, list_values = self.split_value_list(value_list)

        for attr, value in zip_longest(self.writable_single_attrs, single_values):
            if attr is None:
                continue
            attr.value_to(instance, value)
            if value is None and isinstance(attr, DynamicAttribute):
                self.attrs.remove(attr)

        try:
            list_attrs = iter(self.writable_list_attrs)
            next(list_attrs).value_to(instance, list_values)
            for other_list_attr in list_attrs:
                other_list_attr.value_to(instance, None)
        except StopIteration:
            predicate = self.get_or_generate_predicate_obj()
            for value in list_values:
                if isinstance(value, DigitalObject):
                    rel = DigitalObjectRelation.objects.create(
                        subject=instance, predicate=predicate, object=value
                    )
                else:
                    rel = DigitalObjectAttributes.objects.create(
                        subject=instance, predicate=predicate, value=str(value)
                    )
                self.attrs.append(DynamicAttribute(rel))

    def render_new_input(self):
        try:
            list_attrs = iter(self.writable_list_attrs)
            return next(list_attrs).render_empty_input(self.field_name)
        except StopIteration:
            return DigitalObjectT().render_empty_input(self.field_name, {})


@static_attributes.register(DigitalObject)
def digital_object_attributes() -> typing.List[BaseAttribute]:
    vocab = Vocabulary()
    return [
        Attribute(vocab.label, StringT, "label"),
        Attribute(vocab.description, StringT, "description"),
        Attribute(
            AttrObj("Date Created", DCTERMS.created),
            DateT,
            "created",
            is_public=True,
            read_only=True,
        ),
        Attribute(
            AttrObj("Date Modified", SDO.dateModified),
            DateT,
            "last_modified",
            read_only=True,
        ),
        Attribute(
            AttrObj("Date Deleted", SDO.dateDeleted), DateT, "deleted", is_public=True
        ),
        Attribute(
            AttrObj("Owner", FRBR.owner),
            PersonT,
            "owner",
            property_category="Involved People",
            is_public_name="members_is_public",
        ),
        FunctionResultAttribute(
            AttrObj("Members", DCTERMS.creator),
            ListT(PersonT),
            "members",
            property_category="Involved People",
            is_public_name="members_is_public",
        ),
        Attribute(
            AttrObj("Contributors", PRO.relatesToDocument),
            ListT(ContributionT),
            "contributions",
            property_category="Involved People",
            is_public_name="members_is_public",
        ),
        ManyAttribute(
            AttrObj("is part of", DCTERMS.isPartOf),
            ListT(ProjectT),
            "projects",
            recursive=False,
            property_category="Projects",
        ),
    ]


class ImageUrlAttribute(BaseAttribute):
    def __init__(self, **kwargs):
        super().__init__(
            AttrObj("URL", FOAF.primaryTopic),
            UrlT,
            read_only=True,
            is_public_name="download_url_is_public",
            **kwargs,
        )

    def getattr(self, image):
        return f"{settings.HELIPORT_HANDLE_LOCAL_DIGITAL_OBJECT_HOST}{reverse('core:image_data', kwargs={'pk': image.pk})}"


@register_digital_object_attributes(Image)
def image_attributes():
    return [
        TypeAttribute(SemanticDesktop.NFO.Image),
        Attribute(
            AttrObj("mime type", SemanticDesktop.NIE.mimeType), StringT, "mimetype"
        ),
        Attribute(
            AttrObj("Filename", SemanticDesktop.NFO.fileName), StringT, "filename"
        ),
        ImageUrlAttribute(),
    ]


@register_digital_object_attributes(Project)
def project_attributes():
    vocab = Vocabulary()
    return [
        TypeAttribute(vocab.Project),
        ManyAttribute(
            AttrObj("has Part", DCTERMS.hasPart),
            DigitalObjectT,
            "parts",
            casted=True,
            is_public_name="parts_is_public",
            property_category="Related",
        ),
    ]


def get_static_attributes(cls):
    if cls in static_attributes:
        return static_attributes[cls]()
    elif issubclass(cls, DigitalObject):
        return static_attributes[DigitalObject]()
    else:
        return []


def get_dynamic_attributes(obj):
    return [
        DynamicAttribute(attr, property_category="Attributes")
        for attr in obj.attributes.order_by("pk")
    ] + [
        DynamicAttribute(rel, property_category="Relations")
        for rel in obj.relations.order_by("pk")
    ]


def get_all_attributes(obj):
    return get_static_attributes(type(obj)) + get_dynamic_attributes(obj)


def get_all_attributes_bound(obj):
    return [attr.bind_to(obj) for attr in get_all_attributes(obj)]


def get_attributes_by_category(obj, user):
    user_has_read_access = has_read_permission(user, obj)
    result = defaultdict(list)
    for attr in get_all_attributes(obj):
        if attr.exists_on(obj) and (attr.get_is_public(obj) or user_has_read_access):
            result[attr.property_category].append(attr.bind_to(obj))
    return result


def get_attributes_as(user, obj):
    read_access = has_read_permission(user, obj)
    return [
        attr
        for attr in get_all_attributes(obj)
        if read_access or attr.get_is_public(obj)
    ]


def get_bound_attributes_as(user, obj):
    return [attr.bind_to(obj) for attr in get_attributes_as(user, obj)]


def types(obj, bound_attributes=False):
    results = []
    for attribute in get_all_attributes(obj):
        if str(attribute.persistent_id) == str(RDF.type) and attribute.exists_on(obj):
            if bound_attributes:
                results.append(BoundAttribute(attribute, obj))
            else:
                results.append(attribute.value_from(obj))
    return results


class DATACITESerializer:
    @abstractmethod
    def resource_type(self, object_to_serialize):
        pass

    @abstractmethod
    def related_identifiers(self, object_to_serialize, user):
        pass

    @abstractmethod
    def as_dict(self, object_to_serialize, user):
        pass

    def as_json(self, object_to_serialize, user):
        return json.dumps(self.as_dict(object_to_serialize, user))

    def resource_type_dict(self, object_to_serialize):
        concrete_type, general_type = self.resource_type(object_to_serialize)
        assert general_type in [
            "Audiovisual",
            "Collection",
            "DataPaper",
            "Dataset",
            "Event",
            "Image",
            "InteractiveResource",
            "Model",
            "PhysicalObject",
            "Service",
            "Software",
            "Sound",
            "Text",
            "Workflow",
            "Other",
        ]

        return {"resourceType": concrete_type, "resourceTypeGeneral": general_type}


@datacite_serializers.register(DigitalObject)
class DigitalObjectDATACITESerializer(DATACITESerializer):
    """https://github.com/datacite/schema/blob/master/source/json/kernel-4.3/datacite_4.3_schema.json"""

    def resource_type(self, object_to_serialize):
        return "digital object", "Other"

    def related_identifiers(self, object_to_serialize, user):
        if (
            has_read_permission(user, object_to_serialize)
            or object_to_serialize.projects_is_public
        ):
            return [
                {
                    "relatedIdentifier": project.persistent_id,
                    "relatedIdentifierType": project.identifier_scheme.datacite_id,
                    "relationType": "IsPartOf",
                }
                for project in object_to_serialize.projects.all()
                if project.persistent_id
            ]
        else:
            return []

    def as_dict(self, digital_object, user):
        read_access = has_read_permission(user, digital_object)
        creators = []
        for creator in digital_object.members():
            creator_dict = {
                "nameType": "Personal",
                "name": creator.display_name,
            }
            if creator.orcid:
                creator_dict["nameIdentifiers"] = [
                    {
                        "nameIdentifier": creator.orcid,
                        "nameIdentifierScheme": "ORCID",
                        "schemeURI": "http://orcid.org",
                    }
                ]
            creators.append(creator_dict)

        contributors = []
        for contribution in Contribution.objects.filter(contribution_to=digital_object):
            contributor = contribution.contributor
            contribution_dict = {
                "contributionType": contribution.type,
                "name": contributor.display_name,
                "nameType": "Personal",
            }
            if contributor.affiliation:
                contribution_dict["affiliations"] = [
                    {"affiliation": contributor.affiliation}
                ]
            if contributor.orcid:
                contribution_dict["nameIdentifiers"] = [
                    {
                        "nameIdentifier": contributor.orcid,
                        "nameIdentifierScheme": "ORCID",
                        "schemeURI": "http://orcid.org",
                    }
                ]
            contributors.append(contribution_dict)

        identifier = digital_object.persistent_id
        if not identifier:
            identifier = digital_object.full_url()
        identifier_scheme = digital_object.identifier_scheme
        result = {
            "identifiers": [
                {
                    "identifier": identifier,
                    "identifierType": identifier_scheme.datacite_id,
                }
            ],
            "dates": [
                {
                    "dateType": "Created",
                    "date": digital_object.created.strftime("%Y-%m-%d"),
                }
            ],
            "types": self.resource_type_dict(digital_object),
            "schemaVersion": "http://datacite.org/schema/kernel-4",
            "publisher": "HELIPORT",
            "publicationYear": str(timezone.now().year),
        }
        if read_access or digital_object.members_is_public:
            result["creators"] = creators + contributors
        if read_access or digital_object.label_is_public:
            result["titles"] = [{"title": digital_object.label, "lang": "en"}]
        if read_access or digital_object.description_is_public:
            result["descriptions"] = [
                {
                    "description": digital_object.description,
                    "descriptionType": "Abstract",
                    "lang": "en",
                }
            ]

        related_identifiers = self.related_identifiers(digital_object, user)
        if related_identifiers:
            result["relatedIdentifiers"] = related_identifiers

        return result


@datacite_serializers.register(Project)
class ProjectDATACITESerializer(DigitalObjectDATACITESerializer):
    def related_identifiers(self, object_to_serialize, user):
        related = super().related_identifiers(object_to_serialize, user)
        if (
            has_read_permission(user, object_to_serialize)
            or object_to_serialize.parts_is_public
        ):
            for digital_object in DigitalObject.objects.filter(
                projects=object_to_serialize, persistent_id__isnull=False
            ):
                if not digital_object.persistent_id:
                    continue
                related.append(
                    {
                        "relatedIdentifier": digital_object.persistent_id,
                        "relatedIdentifierType": digital_object.identifier_scheme.datacite_id,
                        "relationType": "HasPart",
                    }
                )
        return related

    def resource_type(self, object_to_serialize):
        return "project", "Collection"


@datacite_serializers.register(Image)
class ImageDATACITESerializer(DigitalObjectDATACITESerializer):
    def resource_type(self, image):
        return image.mimetype, "Image"


def to_datacite(obj, user):
    if type(obj) in datacite_serializers:
        serializer = datacite_serializers[type(obj)]()
    elif isinstance(obj, DigitalObject):
        serializer = datacite_serializers[DigitalObject]()
    else:
        raise TypeError(f"No DATACITE serializer registered for {obj}")
    return serializer.as_dict(obj, user)


class RDFSerializer:
    @abstractmethod
    def triples(self, the_object, user):
        pass


@rdf_serializers.register(AttrObj)
class AttrObjRDFSerializer(RDFSerializer):
    def triples(self, the_object, user):
        yield from ()


@rdf_serializers.register(DigitalObject)
class DigitalObjectRDFSerializer(RDFSerializer):
    @staticmethod
    def static_triples(digital_object, user):
        for attr in get_static_attributes(type(digital_object)):
            if has_read_permission(user, digital_object) or attr.get_is_public(
                digital_object
            ):
                yield from attr.rdf_triples(digital_object)

    @staticmethod
    def dynamic_triples(digital_object, user):
        for rel in DigitalObjectRelation.objects.filter(subject=digital_object):
            if rel.is_public or has_read_permission(user, digital_object):
                yield RelationTriple(digital_object, rel.predicate, rel.object)

        for attr in DigitalObjectAttributes.objects.filter(subject=digital_object):
            if attr.is_public or has_read_permission(user, digital_object):
                yield LiteralTriple(digital_object, attr.predicate, attr.value)

    def triples(self, digital_object, user):
        yield from self.static_triples(digital_object, user)
        yield from self.dynamic_triples(digital_object, user)


@rdf_serializers.register(HeliportUser)
class UserRDFSerializer(RDFSerializer):
    def triples(self, user, user_that_wants_to_know):
        if user.display_name:
            yield LiteralTriple(user, FOAF.name, user.display_name)

        if user.auth_user is not None:
            auth_user = user.auth_user
            if auth_user.first_name:
                yield LiteralTriple(user, FOAF.firstName, auth_user.first_name)
            if auth_user.last_name:
                yield LiteralTriple(user, FOAF.lastName, auth_user.last_name)
            if auth_user.email:
                mail_uri = f"mailto:{auth_user.email}"
                if user.email_is_public:
                    yield RelationTriple(user, FOAF.mbox, URIRef(mail_uri))
                if user.email_sha1_is_public:
                    sha1sum = hashlib.sha1(mail_uri.encode())
                    yield LiteralTriple(user, FOAF.mbox_sha1sum, sha1sum.hexdigest())

        if user.affiliation:
            affiliation_node = BNode()
            organization_node = BNode(
                remove_special_chars(
                    f"THE ORGANIZATION FOR AFFILIATION {user.affiliation}"
                )
            )
            yield RelationTriple(user, PRO.holdsRoleInTime, affiliation_node)
            yield RelationTriple(affiliation_node, RDF.type, PRO.RoleInTime)
            yield RelationTriple(affiliation_node, PRO.withRole, SCORO.affiliate)
            yield RelationTriple(
                affiliation_node, PRO.relatesToOrganization, organization_node
            )
            yield RelationTriple(organization_node, RDF.type, FOAF.Organization)
            yield LiteralTriple(organization_node, FOAF.name, user.affiliation)


@rdf_serializers.register(Contribution)
class ContributionRDFSerializer(RDFSerializer):
    def triples(self, contribution, user):
        t = Contribution.ContributionTypes
        yield RelationTriple(
            contribution.contribution_to,
            DCTERMS.contributor,
            contribution.contributor,
        )
        if contribution.type == t.RESEARCH_GROUP:
            yield RelationTriple(
                contribution.contributor, RDF.type, FRAPO.ResearchGroup
            )
        elif contribution.type not in self.type2role():
            yield RelationTriple(contribution.contributor, RDF.type, FOAF.Person)
        else:
            yield RelationTriple(
                contribution.contributor,
                PRO.holdsRoleInTime,
                contribution,
            )
            yield RelationTriple(contribution, RDF.type, PRO.RoleInTime)
            yield RelationTriple(
                contribution, PRO.relatesToDocument, contribution.contribution_to
            )
            yield RelationTriple(
                contribution, PRO.withRole, self.type2role()[contribution.type]
            )

    @staticmethod
    def type2role():
        t = Contribution.ContributionTypes
        # http://dx.doi.org/10.6084/m9.figshare.2075356
        return {
            t.CONTACT_PERSON: SCORO["contact-person"],
            t.DATA_COLLECTOR: SCORO["data-creator"],
            t.DATA_CURATOR: SCORO["data-curator"],
            t.DATA_MANAGER: SCORO["data-manager"],
            t.DISTRIBUTOR: PRO.distributor,
            t.EDITOR: PRO.editor,
            t.HOSTING_INSTITUTION: SCORO["host-institution"],
            t.PRODUCER: PRO.producer,
            t.PROJECT_LEADER: SCORO["project-leader"],
            t.PROJECT_MANAGER: SCORO["project-manager"],
            t.PROJECT_MEMBER: SCORO["project-member"],
            t.REGISTRATION_AGENCY: SCORO["registrar"],
            t.REGISTRATION_AUTHORITY: SCORO["registration-authority"],
            t.RESEARCHER: SCORO["researcher"],
            t.RIGHTS_HOLDER: SCORO["rights-holder"],
            t.SPONSOR: SCORO["sponsor"],
            t.SUPERVISOR: SCORO["supervisor"],
            t.WORK_PACKAGE_LEADER: SCORO["workpackage-leader"],
        }


def to_rdf(digital_object, user, max_depth=100):
    serializer_dict = {k: serializer() for k, serializer in rdf_serializers.items()}
    graph = Graph()
    for prefix, namespace in custom_namespaces:
        graph.bind(prefix, namespace)
    done = set()
    todo = [(digital_object, 0)]
    while todo:
        current, depth = todo.pop()
        if current in done or depth > max_depth:
            continue

        if type(current) in serializer_dict:
            serializer = serializer_dict[type(current)]
        elif isinstance(current, DigitalObject):
            serializer = serializer_dict[DigitalObject]
        else:
            raise TypeError(f"No RDFSerializer registered for {current}")
        for triple in serializer.triples(current, user):
            if triple.recursive:
                todo.extend((o, depth + 1) for o in triple.objects())
            triple.add_to_graph(graph)

        done.add(current)

    return graph


@renderers.register("datacite_json")
class DataCiteJSONRenderer:
    name = "Datacite JSON"

    @staticmethod
    def render(obj, user):
        return JsonResponse(to_datacite(obj, user), json_dumps_params={"indent": 2})


@renderers.register("datacite")
class DataCiteRenderer:
    name = "Datacite XML"

    @staticmethod
    def render(obj, user):
        json_datacite = to_datacite(obj, user)
        xml_datacite = schema43.tostring(json_datacite)
        return HttpResponse(
            xml_datacite, content_type="application/vnd.datacite.datacite+xml"
        )


@renderers.register("jsonld")
@renderers.register("json")
class JsonLdRenderer:
    name = "JSON-LD"

    @staticmethod
    def len_sum(d: dict):
        return sum(
            len(potential_list) if isinstance(potential_list, list) else 1
            for potential_list in d.values()
        )

    @classmethod
    def find_id_usages(cls, item, parent, id_usages):
        recursion = None
        if isinstance(item, dict):
            if "@id" in item:
                id_usages[item["@id"]].append((parent, item))
            recursion = item.values()
        if isinstance(item, list):
            recursion = item

        if recursion is not None:
            for i in recursion:
                cls.find_id_usages(i, parent, id_usages)

    @classmethod
    def inline(cls, graph):
        if not isinstance(graph, list):
            return graph
        if not all(isinstance(d, dict) and d.get("@id") is not None for d in graph):
            return graph
        id_usages = defaultdict(list)
        for d in graph:
            for i in d.values():
                cls.find_id_usages(i, d["@id"], id_usages)
        result = []
        parent_relation = {}
        for d in sorted(graph, key=cls.len_sum):
            d_id = d.get("@id")
            usages = id_usages[d_id]
            if "@id" in d and len(usages) == 1:
                parent, usage = usages[0]
                while parent in parent_relation:
                    parent = parent_relation[parent]
                if d_id != parent:
                    assert usage == {"@id": d_id}, "unexpected situation"
                    for k, v in d.items():
                        usage[k] = v
                    parent_relation[d_id] = parent
                else:
                    result.append(d)
            else:
                result.append(d)

        return result

    @staticmethod
    def filter_prefixes(graph, prefixes):
        min_length = min(len(p) for p in prefixes.values())
        index = defaultdict(list)
        for key, prefix in prefixes.items():
            index[prefix[:min_length]].append((key, prefix))
        prefix_index = dict(index)
        confirmed_prefixes = {}

        for triple in graph:
            for node in triple:
                if isinstance(node, Literal) and isinstance(node.datatype, str):
                    node = node.datatype
                prefix_list = prefix_index.get(node[:min_length], [])
                for key, prefix in list(prefix_list):
                    if node.startswith(prefix):
                        confirmed_prefixes[key] = prefix
                        prefix_list.remove((key, prefix))
        return confirmed_prefixes

    @classmethod
    def render(cls, obj, user):
        graph = to_rdf(obj, user)
        result_json = graph.serialize(format="json-ld", indent=2, auto_compact=True)

        d = json.loads(result_json)
        if set(d) == {"@context", "@graph"}:
            d["@graph"] = cls.inline(d["@graph"])
            if isinstance(d["@context"], dict):
                d["@context"] = cls.filter_prefixes(graph, d["@context"])
        result_json = json.dumps(d, indent=2)

        return HttpResponse(
            result_json,
            content_type="application/ld+json",
        )


@renderers.register("turtle")
class TurtleRenderer:
    name = "Turtle"

    @staticmethod
    def render(obj, user):
        graph = to_rdf(obj, user)
        return HttpResponse(graph.serialize(), content_type="text/turtle")


@renderers.register("xml")
class RdfXmlRenderer:
    name = "RDF XML"

    @staticmethod
    def render(obj, user):
        is_landing_page = re.compile("(.*)/object/([0-9]+)/?")

        def clean(node):
            match = is_landing_page.match(node)
            if match is not None:
                return URIRef(f"{match.group(1)}/object/H{match.group(2)}")
            elif isinstance(node, URIRef) and node.endswith("/"):
                return URIRef(node[:-1])
            else:
                return node

        graph = to_rdf(obj, user)
        clean_graph = Graph()
        for s, p, o in graph:
            clean_graph.add((clean(s), clean(p), clean(o)))
        return HttpResponse(
            clean_graph.serialize(format="pretty-xml"),
            content_type="application/rdf+xml",
        )


@renderers.register("nt")
class TripleRenderer:
    name = "N-Triples"

    @staticmethod
    def render(obj, user):
        graph = to_rdf(obj, user)
        return HttpResponse(
            graph.serialize(format="nt"), content_type="application/n-triples"
        )


@renderers.register("landing_page")
class LandingPageRenderer:
    name = "Default Landing Page"

    @classmethod
    def render(cls, obj, user):
        raise AssertionError("Please handle this renderer explicitly in your view")


@renderers.register("html")
class HtmlRenderer:
    name = "Specific Landing Page"

    @classmethod
    def render(cls, obj, user):
        raise AssertionError("Please handle this renderer explicitly in your view")


def get_unique_renderers(obj_for_urls=None):
    # note that, if multiple renderers are equal, this selects the format
    # that was registered last (the topmost decorator)
    unique_renderers = {r: f for f, r in renderers.items()}
    unique_renderers_by_format = {f: r for r, f in unique_renderers.items()}

    if settings.HELIPORT_HANDLE_RESOLVE_PID_FORMAT_TO and obj_for_urls is not None:
        return {
            f: {
                "name": r.name,
                "url": settings.HELIPORT_HANDLE_RESOLVE_PID_FORMAT_TO.format(
                    format=f, object_id=obj_for_urls.digital_object_id
                ),
            }
            for f, r in unique_renderers_by_format.items()
        }
    else:
        return unique_renderers_by_format


def select_renderer(format_code, accept_header):
    if not format_code:
        format_dict = {
            "application/ld+json": "jsonld",
            "application/vnd.datacite.datacite+json": "datacite_json",
            "application/vnd.datacite.datacite+xml": "datacite",
            "application/rdf+xml": "xml",
            "application/n-triples": "nt",
            "text/turtle": "turtle",
            "text/html": "landing_page",
            "application/xml": "xml",
            "text/xml": "xml",
            "application/json": "json",
            "text/plain": "turtle",
        }
        if accept_header is None:
            accept_header = "*/*"
        mime_type = get_best_match(accept_header, format_dict)
        try:
            format_code = format_dict[mime_type]
        except KeyError:
            return None, f"supported mime types are: {list(format_dict)}"

    try:
        result = renderers[format_code]
    except KeyError:
        return None, f"supported formats are: {list(renderers)}"
    return result, None
