# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.urls import include, path, re_path, reverse_lazy
from django.views.decorators.cache import cache_page
from django.views.generic import RedirectView
from drf_spectacular.views import SpectacularAPIView, SpectacularRedocView
from rest_framework import routers

from .views.api import (
    ContributionViewSet,
    DeleteAllTokensView,
    DigitalObjectAPI,
    GroupViewSet,
    LoginNamedTokenView,
    LogoutTokenView,
    MaintenanceMessageView,
    ProjectViewSet,
    TokenViewSet,
    UserViewSet,
)
from .views.auth import LoginView, LogoutView, RegisterView
from .views.digital_objects import AsDigitalObjectView, DigitalObjectAutocompleteView
from .views.images import ImagesView, ImageView, TakePictureView
from .views.landing_page import (
    LandingPageByPersistentIDView,
    LandingPageView,
    PublicLandingPageView,
)
from .views.project import (
    CreateEmptyProjectView,
    CreateProjectView,
    ImportFromProjectView,
    OpenInProjectView,
    ProjectGraphView,
    ProjectListView,
    ProjectTimelineView,
    ProjectUpdateView,
    ProjectView,
    SubprojectsView,
    get_project_list,
    get_users_autocomplete,
)
from .views.search import SearchBaseView, SearchView
from .views.tags import TagAutocompleteView, TagJsonView, TagsView
from .views.user import (
    LoginInfoUpdateView,
    LoginInfoView,
    RemoteServerLoginView,
    RemoteServerLogoutView,
    UserSettingsView,
)

app_name = "core"
urlpatterns = [
    path("", RedirectView.as_view(url=reverse_lazy("core:project_list")), name="index"),
    re_path(
        "about/(?P<path>.*)",
        RedirectView.as_view(url=f"{settings.HELIPORT_CORE_ABOUT_BASE_URL}/%(path)s"),
        name="about",
    ),
    path("login/", LoginView.as_view(), name="login"),
    path("register/", RegisterView.as_view(), name="register"),
    path("logout/", LogoutView.as_view(), name="logout"),
    path("user/settings/", UserSettingsView.as_view(), name="user_settings"),
    path(
        "user/logout_all_tokens/",
        DeleteAllTokensView.as_view(),
        name="logout_all_tokens",
    ),
    path("user/logins/list/", LoginInfoView.as_view(), name="user_login_list"),
    path(
        "user/logins/update/<int:pk>/",
        LoginInfoUpdateView.as_view(),
        name="user_login_update",
    ),
    path(
        "user/logins/remote_server_login/<int:pk>/",
        RemoteServerLoginView.as_view(),
        name="remote_server_login",
    ),
    path(
        "user/logins/remote_server_logout/<int:pk>/",
        RemoteServerLogoutView.as_view(),
        name="remote_server_logout",
    ),
    path(
        "digital_object_autocomplete/",
        DigitalObjectAutocompleteView.as_view(),
        name="digital_object_autocomplete",
    ),
    path(
        "public/object/H<int:pk>/",
        PublicLandingPageView.as_view(),
        name="public_landing_page",
    ),
    path("object/<int:pk>/", LandingPageView.as_view(), name="landing_page"),
    # required by xml serialization:
    path("object/H<int:pk>/", LandingPageView.as_view(), name="prefixed_landing_page"),
    path(
        "object/",
        LandingPageByPersistentIDView.as_view(),
        name="digital_object_by_persistent_id",
    ),
    path("project/<int:project>/images/", ImagesView.as_view(), name="image_list"),
    path(
        "project/<int:project>/images/take/",
        TakePictureView.as_view(),
        name="take_image",
    ),
    path(
        "images/<int:pk>/data/",
        cache_page(4 * 24 * 60 * 60, cache="image_cache")(ImageView.as_view()),
        name="image_data",
    ),
    path("project/<int:project>/tags/", TagsView.as_view(), name="tag_list"),
    path("project/<int:project>/tags/<int:pk>/", TagsView.as_view(), name="tag_update"),
    path(
        "project/<int:project>/tags/json/",
        TagJsonView.as_view(),
        name="tag_json",
    ),
    path(
        "project/<int:project>/tags/autocomplete/",
        TagAutocompleteView.as_view(),
        name="tag_autocomplete",
    ),
    path(
        "project/<int:project>/subprojects/",
        SubprojectsView.as_view(),
        name="subprojects",
    ),
    path(
        "project/<int:project>/import/",
        ImportFromProjectView.as_view(),
        name="import_project",
    ),
    path("project/list/", ProjectListView.as_view(), name="project_list"),
    path(
        "project/create/empty/",
        CreateEmptyProjectView.as_view(),
        name="project_create_empty",
    ),
    path("project/create/", CreateProjectView.as_view(), name="project_create"),
    path(
        "project/<int:project>/graph/", ProjectGraphView.as_view(), name="project_graph"
    ),
    path(
        "project/<int:project>/timeline/",
        ProjectTimelineView.as_view(),
        name="project_timeline",
    ),
    path("project/<int:project>/", ProjectView.as_view(), name="project"),
    path(
        "project/update/<int:project>/",
        ProjectUpdateView.as_view(),
        name="project_update",
    ),
    path("project/<int:project>/autocomplete_user/", get_users_autocomplete),
    path(
        "project/dropdown_project_list/", get_project_list, name="dropdown_project_list"
    ),
    path("search_base/", SearchBaseView.as_view(), name="search_base"),
    path("search/", SearchView.as_view(), name="search"),
    path(
        "project/<int:project>/as_digital_object/",
        AsDigitalObjectView.as_view(),
        name="as_digital_object",
    ),
    path(
        "project/<int:project>/redirect/",
        OpenInProjectView.as_view(),
        name="open_in_project",
    ),
    re_path(
        r"^favicon.ico$",
        RedirectView.as_view(
            url=staticfiles_storage.url("core/img/favicon.ico"), permanent=True
        ),
    ),
]

# for REST API:
router = routers.DefaultRouter()
router.register(r"projects", ProjectViewSet, basename="projects")
router.register(r"users", UserViewSet, basename="users")
router.register(r"groups", GroupViewSet, basename="groups")
router.register(r"contributions", ContributionViewSet, basename="contributions")
router.register(r"tokens", TokenViewSet, basename="tokens")
router.register(r"digital_object_api", DigitalObjectAPI, basename="digital_object_api")

urlpatterns += [
    path("api-root/", router.get_api_root_view(), name="api_root"),
    path("api/", include(router.urls)),
    path("api/auth/delete_token/", LogoutTokenView.as_view(), name="delete_token"),
    path("api/auth/login_named/", LoginNamedTokenView.as_view(), name="login_named"),
    path("api/auth/", include("knox.urls")),
    path(
        "api/maintenance_message/",
        MaintenanceMessageView.as_view(),
        name="maintenance_message",
    ),
    path("openapi/", SpectacularAPIView.as_view(), name="openapi-schema"),
    path(
        "redoc/",
        SpectacularRedocView.as_view(url_name="core:openapi-schema"),
        name="redoc",
    ),
]
