# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import hashlib
import re
import string
import uuid
from collections import namedtuple

import requests
from django.conf import settings


def generate_object_id():
    return uuid.uuid4()


def remove_special_chars(raw_id):
    result = re.sub("[^A-Za-z0-9]", "X", str(raw_id))
    if len(result) == 0:
        result = "X"
    if result[0] not in string.ascii_letters:
        result = f"X{result}"
    return result


def generate_project_id(user, group, date):
    # create a hash from username and datetime
    hash_base = f"{user}, {date}"
    hash_str = str(int(hashlib.sha256(hash_base.encode()).hexdigest(), 16) % (10**6))
    group_str = ""
    if group is not None:
        group_str = f".{group}"
    return f"{settings.HELIPORT_CORE_HZDR_ID_PREFIX}{group_str}.{date.year}.{hash_str}"


IdentifierScheme = namedtuple("IdentifierScheme", "prefix datacite_id label")
default_identifier_scheme = IdentifierScheme(
    "", "local-resource-identifier-scheme", "Identifier"
)

identifier_schemes = [
    IdentifierScheme("https://doi.org/", "doi", "DOI"),
    IdentifierScheme("https://hdl.handle.net/", "handle", "Handle"),
    IdentifierScheme("https://n2t.net/ark:", "ark", "ark:"),
    IdentifierScheme(
        "https://opencitations.net/oci/", "oci", "Open Citation Identifier"
    ),
    IdentifierScheme("http://purl.org/", "purl", "purl"),
    IdentifierScheme("https://d-nb.info/", "dnb", "DNB catalogue number"),
    IdentifierScheme("https://www.wikidata.org/wiki/", "wikidata", "Wikidata"),
    IdentifierScheme(
        "https://d-nb.info/gnd/", "gnd", "Gemeinsame Normdatei identifier"
    ),
    IdentifierScheme(
        "https://id.loc.gov/authorities/",
        "loc",
        "Library of Congress authority identifier",
    ),
    IdentifierScheme("https://www.worldcat.org/oclc/", "oclc", "OCLC control number"),
    IdentifierScheme("https://", "url", "URL"),
    IdentifierScheme("http://", "url", "URL"),
]


def find_identifier_scheme(identifier):
    if identifier is None:
        return default_identifier_scheme

    for s in identifier_schemes:
        if identifier.startswith(s.prefix):
            return s
    return default_identifier_scheme


def test_resolvability(identifier):
    try:
        response = requests.head(identifier, timeout=10)
        return response.ok
    except requests.RequestException:
        return False


def is_resolvable(identifier: str):
    return identifier.startswith("http://") or identifier.startswith("https://")


def is_globally_unique(identifier: str):
    return is_resolvable(identifier)
