# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import binascii
import json
import logging
import typing
from base64 import b64decode, b64encode
from datetime import datetime

import requests
from django.conf import settings

from heliport.core.utils.string_tools import remove_prefix

logger = logging.getLogger(__name__)
b64_prefix = "data:application/json;charset=utf-8;base64,"


def register_json_handle(handle_name, object_category, url, local_url=None):
    data_url = generate_handle_json_data_url(
        handle_name, local_url, object_category, url, datetime.now()
    )
    return register_handle(handle_name, data_url)


def generate_handle_json_data_url(
    handle_name, local_url, object_category, url, date_created
):
    data = {
        "id": handle_name,
        "created_on": date_created.strftime("%Y/%m/%d, %H:%M"),
        "instance_of": object_category,
        "description": settings.HELIPORT_HANDLE_DESCRIPTION,
        "url": url,
    }
    if local_url is not None and local_url != url:
        data["local_url"] = local_url
    data_url = b64_prefix + b64encode(json.dumps(data).encode("utf-8")).decode()
    return data_url


class HandleServer:
    @staticmethod
    def do_request(method, url, **kwargs):
        return method(url, **kwargs)

    def handle_server_request(self, method, url, json_data=..., authenticate=False):
        headers = {"Content-Type": "application/json"}
        if authenticate:
            headers["Authorization"] = 'Handle clientCert="true"'

        kwargs = {"headers": headers}
        if authenticate:
            kwargs["cert"] = (
                settings.HELIPORT_HANDLE_CERTIFICATE_PATH,
                settings.HELIPORT_HANDLE_PRIVATE_KEY_PATH,
            )

        if json_data is not ...:
            kwargs["json"] = json_data

        return self.do_request(method=method, url=url, **kwargs)

    def get(self, url, json_data=...):
        return self.handle_server_request(requests.get, url, json_data)

    def put(self, url, json_data=...):
        return self.handle_server_request(
            requests.put, url, json_data, authenticate=True
        )


def register_handle(handle_name, target_url, server=None):
    if server is None:
        server = HandleServer()

    handle_url = f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/{handle_name}?overwrite=false"
    value = {"format": "string", "value": target_url}
    request_data = {"index": 1, "type": "URL", "data": value}

    answer = server.put(handle_url, request_data)
    if answer.status_code == 409:
        handle_url = f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/?mintNewSuffix=true"
        answer = server.put(handle_url, request_data)

    logger.info(f"created handle {answer.status_code} {answer.json()}")
    return answer


def get_handle_from_response(answer):
    handle = answer.json().get("handle")
    if not handle:
        return None
    return f"https://hdl.handle.net/{handle}"


def get_key_values(handle, server=None) -> typing.List[typing.Tuple[str, str]]:
    if server is None:
        server = HandleServer()

    handle = remove_prefix(handle, "https://hdl.handle.net/")
    if handle.startswith(f"{settings.HELIPORT_HANDLE_PREFIX}/"):
        # same result as with handle proxy but handle proxy caches results
        response = server.get(f"{settings.HELIPORT_HANDLE_API_URL}/{handle}")
    else:
        response = server.get(f"https://hdl.handle.net/api/handles/{handle}")
    if response.ok:
        value_data = response.json()
        if (
            value_data.get("responseCode") == 1
            and value_data.get("handle").lower() == handle.lower()
            and "values" in value_data
        ):
            return [
                (value["type"], str(value["data"]["value"]))
                for value in value_data["values"]
                if value["type"] != "URL"
            ]

    logger.info(f"values of {handle} not found")
    return []


def get_handle_url(handle, server=None):
    if server is None:
        server = HandleServer()
    handle = remove_prefix(handle, "https://hdl.handle.net/")
    response = server.get(f"{settings.HELIPORT_HANDLE_API_URL}/{handle}")
    if not response.ok:
        return None
    value_data = response.json()
    if (
        value_data.get("responseCode") != 1
        or value_data.get("handle").lower() != handle.lower()
        or "values" not in value_data
        or len(value_data["values"]) == 0
    ):
        return None
    for value in value_data["values"]:
        if value.get("type") == "URL":
            return value["data"]["value"]
    return None


def set_key_values(
    handle, key_values: list, server=None, simulate_register_handles_in_tests=False
) -> bool:
    """
    Returns false if this handle was not generated by HELIPORT
    and returns true if the key_values were set successfully
    """
    if server is None:
        server = HandleServer()

    handle = remove_prefix(handle, "https://hdl.handle.net/")
    if (
        not settings.HELIPORT_HANDLE_REGISTER_HANDLES
        and not simulate_register_handles_in_tests
    ):
        return False
    if not handle.startswith(f"{settings.HELIPORT_HANDLE_PREFIX}/"):
        return False
    b64_url = get_handle_url(handle, server)
    if b64_url is None or not b64_url.startswith(b64_prefix):
        return False
    try:
        url = json.loads(b64decode(b64_url[len(b64_prefix) :]).decode("utf-8")).get(
            "url"
        )
    except (json.JSONDecodeError, binascii.Error, UnicodeDecodeError):
        logger.warning(f"error in b64-json handle {handle}")
        return False
    if url is None or not url.startswith(settings.HELIPORT_HANDLE_DIGITAL_OBJECT_HOST):
        return False

    key_values.insert(0, ("URL", b64_url))
    handle_value_data = [
        {
            "index": i + 1,
            "type": remove_prefix(key, ["https://hdl.handle.net/", "https://doi.org/"]),
            "data": {"format": "string", "value": value},
        }
        for i, (key, value) in enumerate(key_values)
    ]
    handle_url = f"{settings.HELIPORT_HANDLE_API_URL}/{handle}"

    answer = server.put(handle_url, json_data=handle_value_data)
    logger.info(f"updated handle: {answer.json()}")
    return True
