# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import typing
from contextlib import contextmanager
from itertools import chain

from django.conf import settings
from django.db.models import QuerySet
from django.shortcuts import get_object_or_404
from django.utils import timezone

from heliport.core.attribute_description import ConstantAttribute
from heliport.core.models import (
    DigitalObject,
    DigitalObjectRelation,
    Group,
    HeliportUser,
    Project,
    Tag,
    Vocabulary,
    assert_relation,
    digital_object_form_label_and_pk,
)
from heliport.core.serializers import static_attributes
from heliport.core.user_logic.user_information import AuthenticationBackendManager
from heliport.core.utils.colors import pick_color_not_in
from heliport.core.utils.IDs import generate_project_id


def create_empty_project(request, authentication_backend_manager=None):
    """
    Create an empty HELIPORT project in the database and return it.
    The following parameters are already set:
    - owner
    - group (department)
    - hzdr_id
    - category (pid namespace)
    - persistent_id (if configured in settings)
    """
    if authentication_backend_manager is None:
        authentication_backend_manager = AuthenticationBackendManager
    auth_user = request.user
    heliport_user = get_object_or_404(HeliportUser, auth_user=auth_user)
    project = Project(owner=heliport_user)

    group_name = authentication_backend_manager().get_group(auth_user)
    if group_name is not None:
        try:
            project.group = Group.objects.filter(display_name=group_name).get()
        except Group.DoesNotExist:
            new_group = Group(display_name=group_name)
            new_group.save()
            project.group = new_group

    project.save()
    project.hzdr_id = generate_project_id(auth_user, project.group_id, project.created)
    group_str = ""
    if project.group is not None:
        group_str = f"/{project.group}"
    project.category_str = f"{settings.HELIPORT_CORE_HZDR_ID_PREFIX}/Projects/{project.created.year}{group_str}/Project"
    project.save()
    project.register_handle_if_applicable(request)

    return project


@contextmanager
def at_least_one_department(default_name="default_group"):
    """
    context with at least one department in the database::

        with at_least_one_department():
            Group.objects.first()  # does not return None

    The department is not deleted after the context is exited.
    """
    g = Group.objects.first()
    if not g:
        g, is_new = Group.objects.get_or_create(display_name=default_name)
    yield g


def get_or_create_from_label(
    label, pk, request, c_category=None, c_type=None, c_project=None, default=None
):
    """
    Get digital object instance by the first applicable method:

        - get by digital object id (if ``pk`` is truthy)
        - get by ``label`` (if ``label`` uniquely identifies the digital object)
        - create new digital object with label (if ``label`` is truthy)
        - return ``default``

    :param label: The label of the digital object (when need)
    :param pk: digital object id of the digital object (always used if valid)
    :param default: the value that is returned if nothing else is applicable
    :param request: needed for registering handle when creating a new object
    :param c_category: Used when creating a new digital object.
        Sets :attr:`core.models.DigitalObject.category_str`.
        Defaults to ``DEFAULT_OBJECT_NAMESPACE`` in settings.
    :param c_type: Used when creating a new digital object.
        Must be a :class:`core.models.DigitalObject` instance.
        No type is set if not provided.
    :param c_project: Used when creating a new digital object.
        Must be a :class:`core.models.Project` instance.
        No project is added if not provided.
    :return: a saved :class:`core.models.DigitalObject` instance or the value of
        the ``default`` parameter.

    Note: This uses :func:`core.models.digital_object_from_label_and_pk`.
    """
    label_object = digital_object_form_label_and_pk(label, pk)
    if label_object is None and not label:
        return default
    if label_object is None:
        label_object = DigitalObject(label=label)
        if c_category is None:
            c_category = settings.DEFAULT_OBJECT_NAMESPACE
        label_object.category_str = c_category
        label_object.save()
        if c_project is not None:
            label_object.projects.add(c_project)
        if c_type is not None:
            assert_relation(label_object, Vocabulary().type, c_type)
        label_object.register_handle_if_applicable(request)
    return label_object


def digital_objects_by_type(
    the_type: typing.Union[int, str, DigitalObject]
) -> typing.List[QuerySet]:
    """
    Returns a list of QuerySets of objects of the_type.
    Each QuerySet is either a filter() of digital objects or all() objects of a model.

    :param the_type: DigitalObject that is the type. int or decimal str queries digital object by digital_object_id ; other string queries by persistent_id
    """
    result = []
    type_attr = Vocabulary().type

    if isinstance(the_type, str) and the_type.isdecimal():
        the_type = int(the_type)

    if isinstance(the_type, int):
        the_type = DigitalObject.objects.get(digital_object_id=the_type)

    if isinstance(the_type, str):
        type_objects = DigitalObject.objects.filter(persistent_id=the_type)
    else:
        type_objects = [the_type]

    dynamic_results = DigitalObject.objects.filter(
        relations__in=DigitalObjectRelation.objects.filter(
            predicate=type_attr, object__in=type_objects
        )
    )
    if dynamic_results.exists():
        result.append(dynamic_results)

    for a_class, attributes in static_attributes.items():
        if issubclass(a_class, DigitalObject) and any(
            isinstance(attr, ConstantAttribute)
            and attr.attribute == type_attr
            and (attr.value in type_objects or str(attr.value) == str(the_type))
            for attr in attributes()
        ):
            result.append(a_class.objects.all())

    return result


class ListOfQuerySets:
    """
    Manage a list of Django query sets. This is useful to query different kinds
    of DigitalObject subclasses simultaneously.
    """

    def __init__(self, list_of_query_sets: typing.List[QuerySet] = None):
        """__init__"""
        self.data = list_of_query_sets
        if self.data is None:
            self.data = []

    def extend(self, list_of_query_sets: typing.List[QuerySet]):
        """
        Add the query sets form a list of django query sets to this ``ListOfQuerySets``
        """
        if isinstance(list_of_query_sets, ListOfQuerySets):
            list_of_query_sets = list_of_query_sets.data
        self.data.extend(list_of_query_sets)

    def all_digital_objects(self):
        """
        bring this ``ListOfQuerySets`` into a state to include all
        :class:`heliport.core.models.DigitalObject` instances
        """
        self.data = [DigitalObject.objects.exclude(label=12334555)]

    def filter(self, *args, **kwargs):
        """filter all query sets"""
        self.data = [q.filter(*args, **kwargs) for q in self.data]

    def as_list(self, max_count=None):
        """
        Get first ``max_count`` many items from the query sets.
        ``max_count`` of ``None`` means all items.
        """
        if max_count is None:
            return list(set(chain(*self.data)))
        else:
            objects = set()
            for q in self.data:
                objects.update(q[:max_count])
                if len(objects) > max_count:
                    break
            return list(objects)[:max_count]


def tags_of(
    obj: DigitalObject, limit_to_project: Project = None
) -> typing.Iterable[Tag]:
    """
    Get the tags some digital object is tagged with.
    The result can be limited to only tags belonging to some project.
    """
    tag_ids = set(
        DigitalObjectRelation.objects.filter(
            subject=obj,
            predicate=Vocabulary().has_tag,
            object__deleted__isnull=True,
        ).values_list("object", flat=True)
    )
    tags = {Tag.objects.filter(digital_object_id=tag_id).first() for tag_id in tag_ids}
    tags.discard(None)
    tags.update(inferred_tags_of(obj, limit_to_project))
    return tags


def inferred_tags_of(obj: DigitalObject, limit_to_project: Project = None):
    """helper function to get tags inferred from metadata attributes"""
    projects = (
        [limit_to_project] if limit_to_project is not None else obj.projects.all()
    )
    if isinstance(obj, Project):
        projects = [*projects, obj]

    for possible_tag in Tag.objects.filter(projects__in=projects, deleted__isnull=True):
        if has_inferred_tag(obj, possible_tag):
            yield possible_tag


def get_or_create_tag(
    project: Project,
    request,
    label: str,
    attribute_label: str = None,
    attribute_pk: typing.Union[str, int] = None,
    value_label: str = None,
    value_pk: typing.Union[str, int] = None,
):
    """
    Create a new :class:`core.models.Tag` or get an existing one having ``label``.
    Uses :func:`get_or_create_from_label` to get/create ``attribute`` and ``value``.
    Tries to generate a ``color`` for the tag that is most different from existing
    colors of tags in the project.

    :param project: project where tag should be created or gotten from
    :param request: is needed to register handles for new tag, attribute or value
    :param label: label to get or create new tag.
    :param attribute_label: label of attribute that is set by setting tag
        (ignored if attribute_pk is valid)
    :param attribute_pk: digital object id of attribute that is set by setting tag.
    :param value_label: label of value that is set as attribute by setting tag
        (ignored if value_pk is valid)
    :param value_pk: digital object id of value that is set as attribute by setting tag.
    """
    tag = (
        Tag.objects.filter(projects=project, label=label, deleted__isnull=True)
        .order_by("-pk")
        .first()
    )
    if tag is not None:
        return tag

    tag = Tag.objects.create(label=label)
    tag.rgb_color = pick_tag_color(project)
    tag.category_str = settings.TAG_NAMESPACE
    tag.attribute = get_tag_attribute(attribute_label, attribute_pk, project, request)
    tag.value = get_tag_value(value_label, value_pk, tag, project, request)
    tag.save()

    tag.projects.add(project)
    tag.register_handle_if_applicable(request)

    return tag


def pick_tag_color(project):
    """helper function to pick a new tag color in the contest of ``project``"""
    return pick_color_not_in(
        [
            t.rgb_color
            for t in Tag.objects.filter(projects=project, deleted__isnull=True)
        ]
    )


def get_tag_attribute(label, pk, project, request):
    """helper function to get the tag attribute from ``label`` or ``pk``"""
    vocab = Vocabulary()
    return get_or_create_from_label(
        label,
        pk,
        request,
        settings.PROPERTY_NAMESPACE,
        vocab.Property,
        project,
        default=vocab.has_tag,
    )


def get_tag_value(label, pk, tag: Tag, project: Project, request):
    """
    helper function to get tag value form ``label`` or ``pk``

    Important: ``tag.attribute`` should already be set. The tag should be saved before
    setting ``tag.value``.
    """
    vocab = Vocabulary()
    return get_or_create_from_label(
        label,
        pk,
        request,
        settings.CLASS_NAMESPACE if tag.attribute == vocab.type else None,
        vocab.Class if tag.attribute == vocab.type else None,
        project,
        default=tag,
    )


def has_inferred_tag(obj: DigitalObject, tag: Tag):
    """
    helper function to test if ``obj`` has metadata that would allow to
    infer being tagged by ``tag``
    """
    return DigitalObjectRelation.objects.filter(
        subject=obj, predicate=tag.attribute, object=tag.value
    ).exists()


def delete_by_id(digital_object_id: typing.Union[int, str]):
    """mark the digital object with specified id as deleted"""
    obj = get_object_or_404(DigitalObject, digital_object_id=digital_object_id)
    obj.deleted = timezone.now()
    obj.save()
