# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Classes for describing a table. This is used in the :class:`heliport.custom_views.HeliportObjectListView`"""
import typing


class BoundTableAction:
    """
    An Action that is bound to some concrete object; including

    - ``display_name`` - text on a button invoking this action
    - ``action_name`` - unique name to identify this type of action
    - ``style`` - dictionary of information how to display the action
    - ``link`` - concrete target url if this action is a "link" action
    - ``disabled`` - if this action can not be executed

    Bound table actions are usually generated by :meth:`TableDescription.generate_rows`
    via the :meth:`TableAction.bind` function.
    """

    def __init__(self, original_action, link=None, disabled=False):
        self.display_name = original_action.display_name
        self.action_name = original_action.action_name
        self.style = original_action.style
        self.link = link
        self.disabled = disabled

        if self.disabled and self.link is not None:
            self.link = "#"

    @property
    def is_link(self) -> bool:
        """determines if this action takes the user to some web page (is a "link" action)"""
        return self.link is not None

    @property
    def is_external_link(self):
        """
        Determines if this action takes the user to an external web page.
        External web pages are specified starting with the protocol while internal pages are linked to via relative urls
        """
        link = str(self.link)
        return self.is_link and (
            link.startswith("http://") or link.startswith("https://")
        )


class TableAction:
    """
    Action that can be executed on an object represented by a :class:`TableRow`. Properties are:

    - ``display_name`` - text on a button invoking this action
    - ``action_name`` - unique name to identify this action
    - ``style`` - dictionary of information how to display the action
    - ``link_callback`` - function to generate the link if this is a "link" action given an object
    - ``action_callback`` - function to execute the action if this is not a "link" action given an object
    - ``disabled_callback`` - function to return if the action is disabled given an object

    Note that ``link_callback`` is executed before the table is shown
    while ``action_callback`` is executed when user requests the action
    """

    def __init__(
        self,
        display_name,
        action_name,
        link_callback=None,
        action_callback=None,
        disabled_callback=None,
        style=None,
    ):
        self.display_name = display_name
        self.action_name = action_name
        self.link_callback = link_callback
        self.action_callback = action_callback
        self.disabled_callback = disabled_callback
        self.style = style

    def bind(self, obj) -> BoundTableAction:
        """generate a :class:`BoundTableAction` that gets its properties from ``obj``"""

        link = None
        disabled = False

        if self.link_callback is not None:
            link = self.link_callback(obj)
            if link is None:
                link = "#"
                disabled = True
        if self.disabled_callback is not None:
            disabled = self.disabled_callback(obj)
        if self.link_callback is None and self.action_callback is None:
            disabled = True

        return BoundTableAction(self, link, disabled)

    def execute(self, obj):
        """execute this action on ``obj``"""
        return self.action_callback(obj)


class TableColumnDescription:
    """
    A column in a :class:`TableDescription` including

    - ``display_name`` - the table heading
    - ``column_name`` - the name of the attribute where this column gets its values from
    - ``computed_property`` - call this function with the obj instead of getting column_name from the obj
    - ``style`` - dictionary that contains how to present this column

    Note that either ``column_name`` or ``computed_property`` must not be ``None``
    """

    def __init__(self, display_name, column_name, computed_property=None, style=None):
        assert (
            column_name is not None or computed_property is not None
        ), "specify column_name or computed_property"
        self.display_name = display_name
        self.column_name = column_name
        self.style = style
        self.compute_property = lambda obj: getattr(obj, column_name)
        if computed_property is not None:
            self.compute_property = computed_property

    def __repr__(self):
        return f"TableColumnDescription({self.column_name})"


class TableRow:
    """
    A row in a :class:`TableDescription` including

    - ``pk`` - A unique identifier for this row; used to find the relevant object when an action is triggered
    - ``actions`` - List of :class:`BoundTableAction` that can be executed on this row
    - ``data`` - List of values for this row; one value per column
    - ``digital_object_id`` - Use :class:`heliport.models.DigitalObject` specific features; in the :class:`heliport.custom_views.HeliportObjectListView` tags are editable for this DigitalObject

    Rows are generated by :meth:`TableDescription.generate_rows`
    """

    def __init__(
        self,
        pk,
        data: typing.Iterable,
        actions: typing.Iterable[BoundTableAction] = None,
        digital_object_id=None,
    ):
        if actions is None:
            actions = []

        self.pk = pk
        self.actions = list(actions)
        self.data = list(data)
        self.digital_object_id = digital_object_id

    @property
    def has_actions(self):
        """``True`` if this column has any actions"""
        return len(self.actions) > 0


class TableDescription:
    """
    Class for building up the description of a table including

    - ``columns`` - list of :class:`TableColumn`
    - ``actions`` - list of :class:`TableAction`
    - ``data`` - list of objects to generate one row per object
    """

    def __init__(self):
        self.actions = []
        self.columns = []
        self.data = None
        self.dynamic_action_callback: typing.Optional[
            typing.Callable[[typing.Any], typing.List[BoundTableAction]]
        ] = None

    @property
    def has_actions(self):
        """``True`` if this table has any actions"""
        return len(self.actions) > 0

    def generate_rows(self) -> typing.Iterator[TableRow]:
        """
        Generator of :class:`TableRow` s based on ``data``

        :raise ValueError: if this table has no ``data`` yet
        """

        if self.data is None:
            raise ValueError(f"Table has no data! Columns are: {self.columns}")

        for obj in self.data:
            digital_object_id = None
            if hasattr(obj, "digital_object_id"):
                digital_object_id = obj.digital_object_id
            actions = [a.bind(obj) for a in self.actions]
            if self.dynamic_action_callback:
                actions = list(self.dynamic_action_callback(obj)) + actions
            yield TableRow(
                obj.pk,
                [r.compute_property(obj) for r in self.columns],
                actions,
                digital_object_id,
            )

    @property
    def row_list(self) -> typing.List[TableRow]:
        """Same as :meth:`generate_rows` but is a list."""
        return list(self.generate_rows())

    def add_column(self, display_name, row_name, computed_property=None, style=None):
        """
        Append a new column

        :param display_name: column heading
        :param row_name: name of the attribute where the column gets its value from
        :param computed_property: function that computes the value of this column given an object
        :param style: dict with information how to show the column

        :raises AssertionError: if ``row_name`` and ``computed_property`` are ``None``
        """
        self.columns.append(
            TableColumnDescription(
                display_name, row_name, computed_property, style=style
            )
        )

    def remove_column(self, column_name):
        """removes all columns having ``column_name``"""
        self.columns = [c for c in self.columns if c.column_name != column_name]

    def add_action(
        self,
        display_name,
        action_name,
        link_callback=None,
        action_callback=None,
        disabled_callback=None,
        style=None,
    ):
        """
        Append an action

        :param display_name: text on a button invoking this action
        :param action_name: unique name to identify this action
        :param style: dictionary of information how to display the action
        :param link_callback: function to generate the link if this is a "link" action given an object;
            executed before table is shown
        :param action_callback: function to execute the action if this is not a "link" action given an object;
            executed when user requests the action
        :param disabled_callback: function to return if the action is disabled given an object
        """
        self.actions.append(
            TableAction(
                display_name=display_name,
                action_name=action_name,
                link_callback=link_callback,
                action_callback=action_callback,
                disabled_callback=disabled_callback,
                style=style,
            )
        )

    def remove_action(self, action_name):
        """remove all actions that have ``action_name``"""
        self.actions = [a for a in self.actions if a.action_name != action_name]

    def execute(self, action_name, obj):
        """
        Execute an action

        :param action_name: the unique name given to an action when describing the table
        :param obj: the object on which the action should be executed
        :raise ValueError: if ``action_name`` is not a registered name of some action
        """
        for a in self.actions:
            if a.action_name == action_name:
                return a.execute(obj)

        raise ValueError(
            f"{action_name} action was not registered! Columns are: {self.columns}"
        )
