# SPDX-FileCopyrightText: 2023 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.conf import settings
from django.contrib.auth import authenticate, logout
from django.contrib.auth.models import User
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.generic import TemplateView, View

from heliport.core.models import HeliportUser
from heliport.core.user_logic.user_information import AuthenticationBackendManager


class LoginView(TemplateView):
    """Handle main heliport login. Includes showing login screen."""

    #: HTML template rendered by this View
    template_name = "core/user/heliport_login.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template."""
        context = super().get_context_data(**kwargs)
        context["redirect_to"] = self.request.GET.get("next", None)
        context["enable_register"] = settings.AUTHENTICATION_BACKEND_USED == "basic"
        context["is_login_page"] = True
        return context

    def post(self, request):
        """Called by django on http post.

        Show that login failed or log in user. Fetch user information via a concrete
        :class:`heliport.core.user_logic.user_information_managers.managers.UserInformationManager`.
        Handle ``redirect_to`` parameter.
        """
        username = request.POST.get("username")
        password = request.POST.get("password")
        user = authenticate(username=username, password=password)
        if user is None:
            logout(request)
            context = self.get_context_data()
            context["message"] = "Login Failed"
            context["redirect_to"] = request.POST.get("redirect")
            return render(request, self.template_name, context)

        auth_manager = AuthenticationBackendManager()
        auth_manager.fetch_user(user)
        auth_manager.login(user, request)
        return redirect(request.POST.get("redirect", reverse("core:project_list")))


class LogoutView(View):
    """Handle logging out of HELIPORT."""

    def get(self, request):
        """Log out current user. Called by django on http get."""
        logout(request)
        return redirect("core:login")


class RegisterView(TemplateView):
    """Handle registration of new heliport users when using the
    :class:`heliport.core.user_logic.user_information_managers.managers.BasicUserInformationManager`.
    """

    #: HTML template rendered by this View
    template_name = "core/user/register.html"

    def get(self, request, *args, **kwargs):
        """Called by django on http get. Redirect to log in if not using "basic" auth backend setting."""
        if settings.AUTHENTICATION_BACKEND_USED != "basic":
            return redirect("core:login")
        return super().get(request, *args, **kwargs)

    def post(self, request):
        """Create User from provided data. Called by django on http post."""
        if settings.AUTHENTICATION_BACKEND_USED != "basic":
            return redirect("core:login")

        username = request.POST.get("username")
        password = request.POST.get("password")
        password2 = request.POST.get("password2")
        surname = request.POST.get("surname")
        given_name = request.POST.get("given_name")
        email = request.POST.get("email")
        orcid = request.POST.get("orcid")
        message = ""

        if password != password2:
            message = "Passwords don't match"
        if User.objects.filter(username=username).first() is not None:
            message = "Username already exists"
        if message:
            return render(request, self.template_name, context={"message": message})

        user = User.objects.create_user(username, email, password)
        user.first_name = given_name
        user.last_name = surname
        user.save()

        if orcid:
            heliport_user = HeliportUser()
            heliport_user.display_name = (
                AuthenticationBackendManager().get_display_name(user)
            )
            heliport_user.auth_user = user
            heliport_user.set_orcid(orcid)
            heliport_user.save()

        return redirect("core:login")
