# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.contrib.auth.mixins import UserPassesTestMixin
from django.http.response import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render, reverse
from django.utils import timezone
from django.views.generic import TemplateView

from heliport.core.app_interaction import (
    generate_project_dropdown,
    generate_project_navbar,
)
from heliport.core.mixins import HeliportBreadcrumbsMixin, HeliportLoginRequiredMixin
from heliport.core.models import (
    DigitalObject,
    Project,
    Vocabulary,
    digital_object_form_label_and_pk,
)
from heliport.core.utils.exceptions import UserMessage
from heliport.core.utils.form_description import FormDescription
from heliport.core.utils.table_description import (
    BoundTableAction,
    TableAction,
    TableDescription,
)


class HeliportObjectListView(
    HeliportLoginRequiredMixin,
    UserPassesTestMixin,
    HeliportBreadcrumbsMixin,
    TemplateView,
):
    #: Instances of this model are shown in the list and created/edited
    model = DigitalObject
    #: HTML templated rendered by this View
    template_name = "core/base/simple_table_view.html"
    #: Columns shown in the table. Triples of column name, attribute shown in column, column size.
    #: For more control over columns overwrite :meth:`register_columns`.
    #: See implementation of :meth:`register_columns` for definition of column sizes
    columns = [("ID", "pk", "small"), ("Name", "label", "normal")]
    #: Actions that are available for each column. Triples of action name, method name called on the HeliportObjectListView
    #: and style information. For more control over action overwrite :meth:`register_actions`.
    #: See implementation of :meth:`register_actions` for definitions of style information.
    actions = [
        ("Edit", "action_edit", "link_secondary"),
        ("Delete", "action_delete", "danger"),
    ]
    #: Fields editable in this view.
    #: Triple of key name, attribute name ware value is stored and field style
    #: For more control over form fields overwrite :meth:`register_form_columns`.
    #: See implementation of :meth:`register_form_columns` for definitions of field styles.
    edit_fields = [("Name", "label", "normal")]
    #: List of attribute names (second value in triples of :attr:`edit_fields`) that should
    #: be set after ``obj.save()`` is called when creating an ``obj``.
    after_save_attributes = ()
    #: Text displayed as heading
    list_heading = "HELIPORT Object List"
    #: Text displayed as heading over form when creating obj. See also :meth:`get_create_heading`.
    create_heading = None
    #: Text displayed as heading over form when editing obj. See also :meth:`get_update_heading`.
    update_heading = None
    #: Namespace used for handle generation
    category = "HELIPORT/thing"
    #: Url that leads to this view for updating. The "update_url" string is passed to :func:`django.urls.reverse`.
    #: with "project" and "pk" url arguments to get the actual url.
    #: ``update_url`` is required by the default :meth:`action_edit` implementation.
    update_url = None
    #: Url that leads to this view for listing objects. The "update_url" string is passed to :func:`django.urls.reverse`.
    #: with "project" url argument to get the actual url.
    #: ``list_url`` is required by the default :meth:`action_edit` implementation to link back to original page after edditing.
    list_url = None
    #: List of strings that are concatenated and put next to the add button for creating a new object.
    #: See :meth:`utils.form_description.FormDescription.get_add_text` for more information.
    add_text = []
    #: True if handles should be registered for new objects
    register_handle = True
    #: True if tags should be editable in this view
    edit_tags = True

    def __init__(self):
        """Set project to None. It is set to the real project in :meth:`get` and :meth:`post`."""
        super().__init__()
        self.project = None

    def test_func(self):
        """Called by Djangos ``UserPassesTestMixin``. Make sure that user is a project member."""
        user = self.request.user.heliportuser
        return user in self.get_object().members()

    def get_object(self):
        """Return the project for which a list of objects should be shown."""
        return get_object_or_404(Project, pk=self.kwargs["project"])

    def get_update_heading(self):
        """
        Return :attr:`update_heading`. Generate an update heading if this is not specified.
        Generation is done for example based on :attr:`create_heading` by replacing a prefix "Add" or "Create" with "Update".
        """
        if self.update_heading is not None:
            return self.update_heading
        if self.create_heading is not None:
            create_heading = self.create_heading
            create_prefixes = ["Add ", "Create "]
            for p in create_prefixes:
                if create_heading.startswith(p):
                    create_heading = create_heading[len(p) :]
            return f"Update {create_heading}"
        return "Update an Object"

    def get_create_heading(self):
        """
        Return :attr:`create_heading`. Generate a creation heading if this is not specified.
        Generation is done for example based on :attr:`update_heading` by replacing a prefix of "Update" or "Edit" with "Add"
        """
        if self.create_heading is not None:
            return self.create_heading
        if self.update_heading is not None:
            update_heading = self.update_heading
            update_prefixes = ["Update ", "Edit "]
            for p in update_prefixes:
                if update_heading.startswith(p):
                    update_heading = update_heading[len(p) :]
            return f"Add {update_heading}"
        return "Add an Object"

    def get_table(self) -> TableDescription:
        """Helper function for table generation. This calls :meth:`register_columns` and
        :meth:`register_actions` to populate the table description."""
        table = TableDescription()
        self.register_columns(table)
        self.register_actions(table)
        table.dynamic_action_callback = self.dynamic_actions

        table.data = self.model.objects.filter(
            projects=self.project, deleted__isnull=True
        )
        return table

    def get_form(self) -> FormDescription:
        """Helper function for form generation. This calls :meth:`register_form_columns`
        to populate the form description."""
        form = FormDescription()
        self.register_form_columns(form)

        form.add_text = self.add_text
        if "pk" in self.kwargs:
            form.obj = self.get_update_instance(self.kwargs["pk"])
        if self.list_url is not None:
            form.cancel_url = reverse(
                self.list_url, kwargs={"project": self.project.pk}
            )
        return form

    def get_update_instance(self, pk):
        """Returns an instance of :attr:`model` to be updated."""
        return get_object_or_404(self.model, pk=pk)

    def get(self, request, *args, **kwargs):
        """Called by django on http get. Sets ``project`` for which the list should be shown."""
        self.project = self.get_object()
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """
        Called by django to get context when rendering template.
        Collects information like :attr:`list_heading` and generates table (:meth:`get_table`) and
        form (:meth:`get_form`) into context.
        """
        context = super().get_context_data(**kwargs)
        form = self.get_form()
        table = self.get_table()
        context["project"] = self.project
        context["project_dropdown_modules"] = generate_project_dropdown(self.project)
        context["project_navbar_modules"] = generate_project_navbar(self.project)
        context["list_heading"] = self.list_heading
        context["update_heading"] = self.get_update_heading()
        context["create_heading"] = self.get_create_heading()
        context["update"] = "pk" in self.kwargs
        context["edit_tags"] = self.edit_tags
        context["table"] = table
        context["form_description"] = form
        context["vocab"] = Vocabulary()
        if any(c.attr("widget") == "login" for c in form.columns):
            context["logins"] = self.request.user.heliportuser.logininfo_set.all()
        return context

    def dynamic_actions(self, obj):
        if isinstance(obj, DigitalObject):
            casted = obj.casted_or_self()
            actions = casted.actions(self.request.user.heliportuser, self.project)
            return [
                BoundTableAction(TableAction(action.name, "dummy"), action.link)
                for action in actions
            ]
        else:
            return []

    def register_columns(self, table: TableDescription):
        """
        Register the columns on a ``table`` by calling `utils.table_description.TableDescription.add_column`.
        The default implementation uses :attr:`columns` to get information about what columns to add.
        Feel free to implement ``register_columns`` in a subclass.
        """
        for display_name, row_name, column_type in self.columns:
            styles = {
                "small": {"width": 10},
                "normal": {"width": 20},
                "large": {"width": 40},
            }
            table.add_column(display_name, row_name, style=styles[column_type.lower()])

    def register_actions(self, table: TableDescription):
        """
        Register the actions on a ``table`` by calling `utils.table_description.TableDescription.add_action`.
        The default implementation uses :attr:`actions` to get information about what actions to add.
        Feel free to implement ``register_actions`` in a subclass.
        """
        for display_name, my_property, action_type in self.actions:
            callback = getattr(self, my_property)
            types = {
                "link": {"link_callback": callback},
                "link_secondary": {
                    "link_callback": callback,
                    "style": {"class": "btn-secondary"},
                },
                "button": {"action_callback": callback},
                "danger": {
                    "action_callback": callback,
                    "style": {"class": "btn-danger"},
                },
            }
            table.add_action(display_name, my_property, **types[action_type.lower()])

    def register_form_columns(self, form: FormDescription):
        """
        Register the columns on a ``form`` by calling `utils.form_description.FormDescription.add_column`.
        The default implementation uses :attr:`edit_fields` to get information about what columns to add.
        Feel free to implement ``register_form_columns`` in a subclass.
        """
        for display_name, column_name, column_type in self.edit_fields:
            t = column_type.lower()
            if t == "normal":
                form.add_column(display_name, column_name)
            elif t == "large":
                form.add_column(display_name, column_name, attrs={"widget": "textarea"})
            elif t == "color":
                form.add_column(display_name, column_name, attrs={"widget": "color"})
            elif t == "digital_object":
                form.add_column(
                    display_name, column_name, attrs={"widget": "digital_object"}
                )
            else:
                raise ValueError(
                    f"Unsupported type {column_type}! supported types: "
                    "normal, large, color, digital_object"
                )

    def post(self, request, *args, **kwargs):
        """
        Set ``project`` and call the correct one of :meth:`handle_update`, :meth:`handle_create` or :meth:`handle_action`

        Called by django on http post.
        """
        try:
            self.project = self.get_object()

            action_name, target_pk = self.extract_action()
            if action_name is not None and target_pk is not None:
                response = self.handle_action(action_name, target_pk)
            else:
                if "pk" in self.kwargs:
                    response = self.handle_update()
                else:
                    response = self.handle_create()
        except UserMessage as m:
            response = self.handle_error(m.message)

        return response

    def handle_error(self, message):
        """Called by :meth:`post` if a :class:`UserMessage` is raised. It renders the list including the error message"""
        context = self.get_context_data()
        context["message"] = message
        return render(self.request, self.template_name, context)

    def handle_create(self):
        """Handles creation of a new :attr:`model` instance. Including setting attributes."""
        obj = self.get_create_instance()
        form_fields = self.extract_form_fields()
        self.set_update_attributes(obj, form_fields)
        obj.category_str = self.category
        obj.save()
        self.set_update_attributes(obj, form_fields, obj_saved=True)
        obj.projects.add(self.project)
        if self.register_handle:
            obj.register_handle_if_applicable(self.request)
        return HttpResponseRedirect(self.request.path_info)

    def handle_update(self):
        """Handles updating a :attr:`model` instance."""
        obj = self.get_update_instance(self.kwargs["pk"])
        form_fields = self.extract_form_fields()
        self.set_update_attributes(obj, form_fields)
        obj.save()
        self.set_update_attributes(obj, form_fields, obj_saved=True)
        if self.list_url is None:
            raise ValueError("specify list_url to allow sending form data")
        return HttpResponseRedirect(
            reverse(self.list_url, kwargs={"project": self.project.pk})
        )

    def extract_form_fields(self):
        """
        Helper function to extract form fields out of the posted data.
        Data from autocomplete inputs with data separated into a label input and an hidden id input is supported.
        """
        form_fields = {}
        form_data_prefix = "form_data_field_"
        form_object_label_prefix = "form_data_object_label_"
        form_object_id_prefix = "form_data_object_id_"
        for param, value in self.request.POST.items():
            if param.startswith(form_data_prefix):
                form_fields[param[len(form_data_prefix) :]] = value

            if param.startswith(form_object_label_prefix):
                key = param[len(form_object_label_prefix) :]
                object_id = self.request.POST[f"{form_object_id_prefix}{key}"]
                object_label = value
                form_fields[key] = digital_object_form_label_and_pk(
                    object_label, object_id
                )

        return form_fields

    def handle_action(self, action_name: str, target_pk: str):
        """Dispatches an action on a table row."""
        obj = get_object_or_404(self.model, pk=target_pk)
        table = self.get_table()
        response = table.execute(action_name, obj)

        if response is None:
            response = HttpResponseRedirect(self.request.path_info)

        return response

    def extract_action(self):
        """Helper function to extract the action to be dispatched from POST data."""
        action = None
        pk = None
        prefix = "action_trigger_"
        for param, value in self.request.POST.items():
            if param.startswith(prefix):
                action = param[len(prefix) :]
                pk = value
                break
        return action, pk

    def get_create_instance(self):
        """Create a new instance of :attr:`model`."""
        return self.model()

    def set_update_attributes(self, obj, form_fields, obj_saved=False):
        """
        Is called by :meth:`handle_create` and :meth:`handle_update` after the ``obj`` is saved.
        The default implementation sets all attributes in :attr:`after_save_attributes`.
        """
        for key, value in form_fields.items():
            if (key in self.after_save_attributes) == obj_saved:
                setattr(obj, key, value)

    def action_edit(self, obj):
        """
        Implements the edit action so that the string "action_edit" can be used in :attr:`actions`.
        The edit action is a link action that links to :attr:`update_url`, which points normally to this view itself.
        """
        if self.list_url is None:
            raise ValueError(
                "the default editing function requires the list_url to be specified"
            )
        return reverse(
            self.update_url, kwargs={"project": self.project.pk, "pk": obj.pk}
        )

    def action_delete(self, obj):
        """
        Implements the delete action so that the string "action_delete" can be used in :attr:`actions`.
        The default implementation sets ``deleted`` to the current time and saves the object.
        """
        obj.deleted = timezone.now()
        obj.save()
