# SPDX-FileCopyrightText: 2023 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.conf import settings
from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import Http404, redirect
from django.urls import reverse
from django.views.generic import DetailView, RedirectView

from heliport.core.models import DigitalObject
from heliport.core.permissions import has_write_permission
from heliport.core.serializers import (
    HtmlRenderer,
    LandingPageRenderer,
    get_all_attributes,
    get_attributes_by_category,
    get_unique_renderers,
    select_renderer,
)
from heliport.core.utils.string_tools import remove_prefix


class PublicLandingPageView(DetailView):
    """Handle rendering public landing pages for Digital Objects.
    Normal version is :class:`LandingPageView`. This view exists in addition to that view in order to remove
    functionality to reduce probability of showing non-public information.
    Used by `landing page server <https://codebase.helmholtz.cloud/fwcc/data-management/landing-page-server>`_ to make
    landing pages available outside the internal network.
    """

    #: HTML template rendered by this View
    template_name = "core/digital_object/public_landingpage.html"

    #: Django model handled in this view - used by django ``DetailView``
    model = DigitalObject

    def get_context_data(self, **kwargs):
        """Get relevant infos in addition to the Digital Object itself.
        Called by django to get context when rendering html template."""

        context = super().get_context_data(**kwargs)
        obj = self.object.casted_or_self()
        context["renderers"] = get_unique_renderers(obj)
        context["static_html"] = True
        context["attributes_by_category"] = get_attributes_by_category(
            obj, user=None
        ).items()
        context["host"] = self.request.get_host()
        context["protocol"] = self.request.scheme
        context["favicon_url"] = settings.HELIPORT_CORE_PUBLIC_FAVICON_URL
        no_redirect_param = self.request.GET.get("public_version")
        if no_redirect_param is not None and str(no_redirect_param).lower() != "false":
            context["no_redirect"] = True
        return context

    def get(self, request, *args, **kwargs):
        """Select renderer based on ``ACCEPT`` header and render response. Called by django on http get."""

        digital_object = self.get_object().casted_or_self()

        renderer, error = select_renderer(
            request.GET.get("format"), request.META.get("HTTP_ACCEPT")
        )
        if error is not None:
            return HttpResponse(error, status=406, content_type="text/plain")

        if renderer in (LandingPageRenderer, HtmlRenderer):
            return super().get(request, *args, **kwargs)

        response = renderer.render(digital_object, user=None)
        return response


class LandingPageView(DetailView):
    """Handle rendering landing pages in different formats (including html, rdf, datacite) for Digital Objects.
    For only public information :class:`PublicLandingPageView` also exists."""

    #: HTML template rendered by this View
    template_name = "core/digital_object/landingpage.html"
    #: Django model handled in this view - used by django ``DetailView``
    model = DigitalObject

    def get_context_data(self, **kwargs):
        """Get relevant infos in addition to the Digital Object itself.
        Called by django to get context when rendering html template."""
        context = super().get_context_data(**kwargs)
        context["renderers"] = get_unique_renderers()
        obj = self.object.casted_or_self()
        user = None
        if hasattr(self.request.user, "heliportuser"):
            user = self.request.user.heliportuser
        context["attributes_by_category"] = get_attributes_by_category(
            obj, user
        ).items()
        context["has_write_permission"] = has_write_permission(user, obj)
        context["host"] = self.request.get_host()
        context["protocol"] = self.request.scheme
        return context

    @staticmethod
    def get_custom_url(obj, request):
        """Get the url of a special page in heliport for showing ``obj``. This information is defined on ``obj``
        via a ``get_custom_url()`` function. For example a project is shown as the heliport project graph. (see
        :meth:`heliport.core.models.Project.get_custom_view_url`)"""
        user = None
        if hasattr(request.user, "heliportuser"):
            user = request.user.heliportuser

        if hasattr(obj, "get_custom_view_url"):
            return obj.get_custom_view_url(user)
        else:
            return None

    def get(self, request, *args, **kwargs):
        """
        Renderer selection based on `Accept header <https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept>`_
        and handling of special renderers like showing individual page with :meth:`get_custom_url`.

        Called by django on http get.
        """
        digital_object = self.get_object().casted_or_self()
        user = None
        if hasattr(request.user, "heliportuser"):
            user = request.user.heliportuser

        renderer, error = select_renderer(
            request.GET.get("format"), request.META.get("HTTP_ACCEPT")
        )
        if error is not None:
            return HttpResponse(error, status=406, content_type="text/plain")

        if renderer == LandingPageRenderer:
            return super().get(request, *args, **kwargs)
        if renderer == HtmlRenderer:
            custom_url = self.get_custom_url(digital_object, request)
            if custom_url is None:
                return super().get(request, *args, **kwargs)
            else:
                return redirect(custom_url)

        response = renderer.render(digital_object, user)
        return response

    def post(self, request, *args, **kwargs):
        """Handle editing what attributes are publicly shown. Called by django on http post."""
        public_selection = request.POST.getlist("public_attributes")
        obj = self.get_object().casted_or_self()
        user = None
        if hasattr(request.user, "heliportuser"):
            user = request.user.heliportuser
        if not has_write_permission(user, obj):
            return HttpResponseForbidden()
        attributes = {
            attr.get_id(obj): attr.bind_to(obj) for attr in get_all_attributes(obj)
        }

        todo = []
        for attr_is_public_select in public_selection:
            is_public = request.POST.get(attr_is_public_select) == "public"
            attr_id = remove_prefix(attr_is_public_select, "public_attributes_")
            attr = attributes.get(attr_id)
            if attr is not None and attr.is_public != is_public:
                # Not setting is_public directly and checking if something changed
                # helps when two attributes share an is_public boolean
                todo.append((attr, is_public))

        for attr, is_public in todo:
            attr.is_public = is_public

        return redirect("core:landing_page", **kwargs)


class LandingPageByPersistentIDView(RedirectView):
    """
    Handle redirecting to landing page when given query parameters ``persistent_id``, ``handle`` or ``doi``.
    """

    def get_redirect_url(self, *args, **kwargs):
        """Find the digital object and return url of its landing page. Called by ``RedirectView``."""
        persistent_id = self.request.GET.get("persistent_id")
        handle = self.request.GET.get("handle")
        doi = self.request.GET.get("doi")
        if handle:
            persistent_id = f"https://hdl.handle.net/{handle}"
        if doi:
            persistent_id = f"https://doi.org/{doi}"
        if not persistent_id:
            raise Http404('"persistent_id", "handle" or "doi" parameter not specified')

        digital_object = DigitalObject.objects.filter(
            persistent_id__iexact=persistent_id
        ).first()
        if digital_object is None:
            digital_object = DigitalObject.objects.filter(
                generated_persistent_id__iexact=persistent_id
            ).first()
        if digital_object is None:
            raise Http404("persistent id not found")

        url = reverse("core:landing_page", kwargs={"pk": digital_object.pk})
        result_format = self.request.GET.get("format")
        if result_format:
            url = f"{ url }?format={result_format}"
        return url
