# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
from collections import namedtuple
from typing import Any, Dict, List, Union

import yaml

from .cwl_parameter import CwlParameter, ParameterValue


class CwlJob:
    def __init__(
        self,
        parameters: List[CwlParameter],
        job_dict: Dict[str, Union[Any, ParameterValue]] = None,
        prefer_json=False,
        tag_class=None,
    ):
        if job_dict is None:
            job_dict = {}
        self.prefer_json = prefer_json
        self.original_parameters = parameters
        self.tag_class = tag_class

        self.dict = {}
        self.unsupported_dict = {}

        for parameter in parameters:
            p = job_dict.get(parameter.name)
            if parameter.supported:
                self.dict[parameter.name] = parameter.generate_parameter_value(p)
            elif isinstance(p, ParameterValue):
                self.unsupported_dict[parameter.name] = p.cwl
            else:
                self.unsupported_dict[parameter.name] = p

    def get_value(self, name):
        parameter_value = self.dict.get(name)
        if parameter_value is None:
            return None
        else:
            return parameter_value.value

    @property
    def is_valid(self):
        return all(p.is_valid for p in self.dict.values())

    @property
    def cwl(self):
        result = {key: value.cwl for key, value in self.dict.items()}
        result.update(self.unsupported_dict)
        return result

    @property
    def json(self):
        return json.dumps(self.cwl, indent=3)

    @property
    def yaml(self):
        return yaml.safe_dump(self.cwl)

    @property
    def cwl_str(self):
        if self.prefer_json:
            return self.json
        else:
            return self.yaml

    @property
    def unsupported_cwl_str(self):
        if self.prefer_json:
            return json.dumps(self.unsupported_dict)
        else:
            return yaml.safe_dump(self.unsupported_dict)

    @property
    def has_parameters(self):
        return bool(self.dict or self.unsupported_dict)

    @property
    def parameter_details(self):
        details = namedtuple(
            "ParameterDetails",
            [
                "name",
                "description",
                "type",
                "count",
                "string_typed",
                "type_name",
                "value",
                "supported_type",
                "value_list",
                "addable",
                "tag_list",
            ],
        )
        string_typed_types = ["string", "long", "int", "double", "float"]
        name_dict = {
            "string": "String",
            "boolean": "Boolean",
            "long": "Integer",
            "int": "int32",
            "double": "Floating Point Number",
            "float": "float32",
        }
        return [
            details(
                name=p.name,
                description=p.description,
                type=p.type_type,
                count=p.type_count,
                string_typed=p.type_type in string_typed_types,
                type_name=name_dict.get(p.type_type, p.type_type),
                supported_type=p.supported,
                value=self.get_value(p.name),
                value_list=self.as_list(self.get_value(p.name), p.type_count),
                addable=p.type_count == "[]"
                or len(self.as_list(self.get_value(p.name), p.type_count)) == 0,
                tag_list=self.get_tag_list(p.name),
            )
            for p in self.original_parameters
        ]

    @staticmethod
    def as_list(value, count):
        if count == "":
            return [value]
        elif count == "?":
            if value is None:
                return []
            else:
                return [value]
        elif isinstance(value, list):
            return value
        else:
            return []

    def get_tag_list(self, name):
        v = self.dict.get(name)
        if v is None or self.tag_class is None:
            return []
        else:
            result = []
            for tag_id in v.tag_list:
                tag = self.tag_class.objects.filter(pk=tag_id).first()
                if tag is not None:
                    result.append(
                        {
                            "color": tag.html_color,
                            "name": tag.label,
                            "id": tag.pk,
                            "foreground": tag.html_foreground_color,
                        }
                    )
            return result
