# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.urls import reverse

from heliport.core.app_interaction import Module


class ArbitraryCWLModule(Module):
    name = "CWL File"
    module_id = "arbitrary_cwl"

    def get_url(self, project):
        return reverse(
            "cwl_execution:arbitrary_cwl_list", kwargs={"project": project.pk}
        )

    def is_configured(self, project):
        from .models import Executable

        cwl_t = Executable.ExecutableTypes.ARBITRARY_CWL
        return bool(Executable.objects.filter(projects=project, type=cwl_t))


class ToolModule(Module):
    name = "Tool"
    module_id = "tool"

    def get_url(self, project):
        return reverse("cwl_execution:tool_list", kwargs={"project": project.pk})

    def is_configured(self, project):
        from .models import Executable

        cwl_t = Executable.ExecutableTypes.TOOL
        return bool(Executable.objects.filter(projects=project, type=cwl_t))


class WorkflowModule(Module):
    name = "Workflow"
    module_id = "workflow"

    def get_url(self, project):
        return reverse("cwl_execution:workflow_list", kwargs={"project": project.pk})

    def is_configured(self, project):
        from .models import Executable

        cwl_t = Executable.ExecutableTypes.WORKFLOW
        return bool(Executable.objects.filter(projects=project, type=cwl_t))


class JobModule(Module):
    name = "CWL Jobs"
    module_id = "job"

    def get_url(self, project):
        return reverse("cwl_execution:job_list", kwargs={"project": project.pk})

    def is_configured(self, project):
        from .models import Job

        return bool(Job.objects.filter(projects=project))


def get_api_root():
    return reverse("cwl_execution:api_root")


def get_search_url():
    return reverse("cwl_execution:search")


def serialize_project(project):
    from .models import Executable

    executables = Executable.objects.filter(projects=project)
    workflows = []
    tools = []
    custom_cwl = []
    for exe in executables:
        exe_dict = {}
        jobs = []

        exe_dict["heliport:executable_id"] = exe.executable_id
        exe_dict["heliport:cwl"] = exe.cwl
        for job in exe.job_set.all():
            execution_logs = []
            job_dict = {
                "heliport:job_id": job.job_id,
                "rdfs:label": job.label,
                "datacite:hasDescription": job.description,
                "heliport:output_text": repr(job.output_text),
                "heliport:repeat": job.repeat,
                "heliport:interval": f"{job.interval} days",
                "heliport:environment": job.setup_command,
                "heliport:cwl": job.cwl,
            }

            for conf in job.userjobconfig_set.all():
                for log in conf.executionlog_set.all():
                    execution_logs.append(
                        {
                            "time:hasBeginning": str(log.start_time),
                            "time:hasEnd": str(log.end_time),
                            "time:hasDuration": log.duration_str,
                            "heliport:memory_usage": log.memory_usage_dict,
                            "heliport:profiling": {
                                step: f"{time}s" for step, time in log.profiling.items()
                            }
                            if log.profiling
                            else None,
                            "heliport:execution_id": log.execution_id,
                            "heliport:was_aborted": log.aborted,
                        }
                    )
            if execution_logs:
                job_dict["heliport:was_executed"] = execution_logs

            jobs.append(job_dict)

        if jobs:
            exe_dict["heliport:has_Job"] = jobs

        if exe.type == exe.ExecutableTypes.TOOL:
            tools.append(exe_dict)
        elif exe.type == exe.ExecutableTypes.WORKFLOW:
            workflows.append(exe_dict)
        else:
            custom_cwl.append(exe_dict)

    result = {}
    if workflows:
        result["heliport:has_Workflow"] = workflows
    if tools:
        result["heliport:has_Tool"] = tools
    if custom_cwl:
        result["heliport:has_CustomCWL"] = custom_cwl
    return result
