# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

from django.conf import settings
from django.shortcuts import reverse
from rdflib import FOAF, RDF, URIRef
from rest_framework import serializers

from heliport.core.permissions import has_read_permission
from heliport.core.serializers import (
    DigitalObjectDATACITESerializer,
    DigitalObjectRDFSerializer,
    datacite_serializers,
    rdf_serializers,
)
from heliport.core.vocabulary import FABIO
from heliport.core.vocabulary_descriptor import RelationTriple

from .cwl_helpers.cwl2rdf import cwl_to_rdf
from .models import Executable, ExecutionLog, Job, UserJobConfig


class ExecutableSerializer(serializers.ModelSerializer):
    class Meta:
        model = Executable
        fields = ["executable_id", "cwl_type", "projects", "type", "cwl", "cwl_json"]


class JobSerializer(serializers.ModelSerializer):
    class Meta:
        model = Job
        fields = [
            "job_id",
            "projects",
            "label",
            "status",
            "description",
            "output_text",
            "output_code",
            "cwl",
            "repeat",
            "interval",
            "base_executable",
            "setup_command",
        ]


class UserJobConfigSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserJobConfig
        fields = ["user_job_config_id", "job", "ssh_login", "directory"]


class ExecutionLogSerializer(serializers.ModelSerializer):
    class Meta:
        model = ExecutionLog
        fields = [
            "execution_log_id",
            "config",
            "start_time",
            "end_time",
            "memory_usage",
            "execution_id",
            "aborted",
            "step_times",
            "output_code",
            "error_message",
            "user",
        ]


@datacite_serializers.register(Executable)
class ExecutableDATACITESerializer(DigitalObjectDATACITESerializer):
    def resource_type(self, exe):
        return exe.get_type_display(), "Workflow"


@rdf_serializers.register(Executable)
class ExecutableRDFSerializer(DigitalObjectRDFSerializer):
    def static_triples(self, exe, user):
        yield from super().static_triples(exe, user)
        yield RelationTriple(exe, RDF.type, FABIO.Workflow)

        cwl_url = reverse("cwl_execution:exe_cwl_json", kwargs={"pk": exe.pk})
        main_cwl_url = f"{settings.HELIPORT_HANDLE_LOCAL_DIGITAL_OBJECT_HOST}{cwl_url}{'' if cwl_url.endswith('/') else '/'}main"
        if exe.download_url_is_public or has_read_permission(user, exe):
            yield RelationTriple(exe, FOAF.primaryTopic, URIRef(main_cwl_url))
        if exe.cwl_is_public or has_read_permission(user, exe):
            for s, p, o in cwl_to_rdf(exe.cwl, main_cwl_url):
                yield RelationTriple(s, p, o)


@datacite_serializers.register(Job)
class JobDATACITESerializer(DigitalObjectDATACITESerializer):
    def resource_type(self, job):
        return "job", "Workflow"

    def related_identifiers(self, job, user):
        related = super().related_identifiers(job, user)
        exe = job.base_executable
        if exe.persistent_id:
            related.append(
                {
                    "relatedIdentifier": exe.persistent_id,
                    "relatedIdentifierType": exe.identifier_scheme.datacite_id,
                    "relationType": "Requires",
                }
            )
        return related
