# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json

from django.contrib.auth.models import User
from django.shortcuts import reverse
from django.test import TestCase

from heliport.core.models import HeliportUser, Project

from .models import DataManagementPlan


class DataManagementPlanTests(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username="testuser")
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.client.force_login(self.auth_user)
        self.project = Project.objects.create(owner=self.user)

    def test_list(self):
        dmp = DataManagementPlan()
        dmp.save()
        dmp.projects.add(self.project)

        response = self.client.get(
            reverse("data_management_plan:list", kwargs={"project": self.project.pk})
        )
        self.assertEqual(200, response.status_code)
        found_dmp = DataManagementPlan.objects.get(
            pk=dmp.pk, projects=response.context_data["project"]
        )
        self.assertEqual(found_dmp, dmp)

        response = self.client.get(
            reverse(
                "data_management_plan:update",
                kwargs={"project": self.project.pk, "pk": dmp.pk},
            )
        )
        self.assertEqual(200, response.status_code)
        found_dmp = DataManagementPlan.objects.get(
            pk=dmp.pk, projects=response.context_data["project"]
        )
        self.assertEqual(found_dmp, dmp)

    def test_create(self):
        self.client.post(
            reverse("data_management_plan:list", kwargs={"project": self.project.pk}),
            data={
                "description": "data_management321",
                "link": "https://www.google.com",
            },
        )
        test_obj = DataManagementPlan.objects.filter(
            description="data_management321", projects=self.project
        ).first()
        self.assertIsNotNone(test_obj)
        self.assertEqual("https://www.google.com", test_obj.link)

    def test_update(self):
        dmp = DataManagementPlan.objects.create(description="data_management456")
        dmp.link = "http://www.google.com"
        dmp.projects.add(self.project)
        response = self.client.post(
            reverse(
                "data_management_plan:update",
                kwargs={"project": self.project.pk, "pk": dmp.pk},
            ),
            data={"description": "data_plan789", "link": "https://www.bing.com"},
            follow=True,
        )
        self.assertEqual(200, response.status_code)
        dmp = DataManagementPlan.objects.get(pk=dmp.pk)
        self.assertIn(self.project, dmp.projects.all())
        self.assertEqual("data_plan789", dmp.description)
        self.assertEqual("https://www.bing.com", dmp.link)

    def test_delete(self):
        data_plan = DataManagementPlan.objects.create()
        data_plan2 = DataManagementPlan.objects.create()
        data_plan.projects.add(self.project)
        data_plan2.projects.add(self.project)
        response = self.client.post(
            reverse("data_management_plan:list", kwargs={"project": self.project.pk}),
            data={"remove": data_plan2.pk},
            follow=True,
        )
        self.assertEqual(200, response.status_code)
        obj1 = DataManagementPlan.objects.filter(
            pk=data_plan.pk, deleted__isnull=True
        ).first()
        obj2 = DataManagementPlan.objects.filter(
            pk=data_plan2.pk, deleted__isnull=True
        ).first()
        self.assertIsNotNone(obj1)
        self.assertIsNone(obj2)
        self.assertIn(self.project, obj1.projects.all())

    def test_add_protocol_create(self):
        self.client.post(
            reverse("data_management_plan:list", kwargs={"project": self.project.pk}),
            data={"description": "data_management321", "link": "www.google.com"},
        )
        test_obj = DataManagementPlan.objects.filter(
            description="data_management321", projects=self.project
        ).first()
        self.assertIsNotNone(test_obj)
        self.assertEqual("https://www.google.com", test_obj.link)

    def test_add_protocol_update(self):
        dmp, is_new = DataManagementPlan.objects.get_or_create(
            description="data_management456"
        )
        dmp.link = "http://www.google.com"
        dmp.projects.add(self.project)

        response = self.client.post(
            reverse(
                "data_management_plan:update",
                kwargs={"project": self.project.pk, "pk": dmp.pk},
            ),
            data={"description": "data_plan789", "link": "www.bing.com"},
            follow=True,
        )
        dmp = DataManagementPlan.objects.get(pk=dmp.pk)
        self.assertIn(response.context_data["project"], dmp.projects.all())
        self.assertEqual("https://www.bing.com", dmp.link)


class SearchAndAPITest(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username="testuser")
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.project, is_new = Project.objects.get_or_create(owner=self.user)
        self.client.force_login(self.auth_user)

    def test_access_search(self):
        response = self.client.get(reverse("data_management_plan:search"))
        self.assertEqual(200, response.status_code)

    def test_data_management_plan_findable(self):
        dmp = DataManagementPlan.objects.create(description="DMP321")
        dmp.projects.add(self.project)
        response = self.client.get(f'{reverse("data_management_plan:search")}?q=321')
        self.assertContains(response, "DMP321")

    def test_api(self):
        dmp = DataManagementPlan.objects.create(description="DMP654")
        dmp.projects.add(self.project)
        response = self.client.get(reverse("data_management_plan:api_root"))
        self.assertEqual(200, response.status_code)
        urls = json.loads(response.content.decode())
        response = self.client.get(urls["data_management_plan"])
        self.assertEqual(200, response.status_code)
        self.assertContains(response, "DMP654")


class ProjectSerializeTest(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username="testuser")
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.project, is_new = Project.objects.get_or_create(owner=self.user)
        self.client.force_login(self.auth_user)
        self.dmp_description = "DMP TO SERIALIZE"
        self.dmp_link = "https://DMP_LINK"
        self.dmp = DataManagementPlan(
            description=self.dmp_description,
            persistent_id="https://hdl.handle.net/aaa",
        )
        self.dmp.save()
        self.dmp.link = self.dmp_link
        self.dmp.projects.add(self.project)

    @staticmethod
    def landing_page_url(obj):
        return reverse("core:landing_page", kwargs={"pk": obj.digital_object_id})

    def test_json_ld(self):
        response = self.client.get(
            f"{self.landing_page_url(self.project)}?format=jsonld"
        )
        self.assertContains(response, self.dmp_description)
        response = self.client.get(f"{self.landing_page_url(self.dmp)}?format=jsonld")
        metadata = response.json()
        dmps = [
            a
            for a in metadata["@graph"]
            if a.get("@type") == "fabio:DataMangementPlan"  # Typo in ontology.
        ]
        self.assertEqual(1, len(dmps))
        dmp = dmps[0]
        self.assertIn("@id", dmp)
        self.assertEqual(self.dmp_description, dmp["dcterms:description"])
        self.assertContains(response, self.dmp_link)
        self.assertContains(response, "https://hdl.handle.net/aaa")

    def test_datacite(self):
        response = self.client.get(
            f"{self.landing_page_url(self.dmp)}?format=datacite_json"
        )
        metadata = response.json()
        self.assertEqual(
            [
                {
                    "description": self.dmp_description,
                    "descriptionType": "Abstract",
                    "lang": "en",
                }
            ],
            metadata["descriptions"],
        )
        self.assertEqual(
            {"resourceType": "DataManagementPlan", "resourceTypeGeneral": "Text"},
            metadata["types"],
        )

    def test_rest_api(self):
        response = self.client.get("/api/projects/urls/")
        self.assertContains(response, "/data_management_plan/")
        response = self.client.get("/data_management_plan/api/data_management_plan/")
        dmp = response.json()["results"][0]
        self.assertEqual(
            {
                "data_management_plan_id": 1,
                "description": self.dmp_description,
                "link": self.dmp_link,
                "projects": [self.project.pk],
            },
            dmp,
        )
