# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import django_filters
from django.conf import settings
from django.db.models import Q
from django.shortcuts import get_object_or_404, redirect, reverse
from django.utils import timezone
from django.views.generic import CreateView, TemplateView, UpdateView
from rest_framework import filters as rest_filters

from heliport.core.mixins import (
    HeliportLoginRequiredMixin,
    HeliportObjectMixin,
    HeliportProjectMixin,
)
from heliport.core.models import Project
from heliport.core.permissions import is_object_member
from heliport.core.views import HeliportModelViewSet

from .models import DataManagementPlan
from .serializers import DataManagementPlanSerializer


class DataManagementPlanView(HeliportProjectMixin, CreateView):
    model = DataManagementPlan
    template_name = "data_management_plan/data_management_plan.html"
    fields = ["description"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["plans"] = DataManagementPlan.objects.filter(
            projects=context["project"], deleted__isnull=True
        )
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        data_management_plan_id = request.POST.get("remove")
        if data_management_plan_id:
            data_management_plan = get_object_or_404(
                DataManagementPlan, pk=data_management_plan_id
            )
            data_management_plan.deleted = timezone.now()
            data_management_plan.save()
            return redirect(reverse("data_management_plan:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        form.instance.category_str = settings.DATA_MANAGEMENT_PLAN_NAMESPACE
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.save()
        form.instance.link = self.request.POST.get("link")
        form.instance.projects.add(project)
        form.instance.register_handle_if_applicable(self.request)
        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_management_plan:list", kwargs=self.kwargs)


class DataManagementPlanUpdateView(HeliportObjectMixin, UpdateView):
    model = DataManagementPlan
    template_name = "data_management_plan/data_management_plan.html"
    fields = ["description"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["plans"] = DataManagementPlan.objects.filter(
            projects=context["project"], deleted__isnull=True
        )
        context["update"] = True
        return context

    def form_valid(self, form):
        form.instance.save()
        form.instance.link = self.request.POST.get("link")
        return super().form_valid(form)

    def get_success_url(self):
        return reverse(
            "data_management_plan:list", kwargs={"project": self.kwargs["project"]}
        )


class SearchView(HeliportLoginRequiredMixin, TemplateView):
    template_name = "data_management_plan/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get("q", "")
        context["query"] = query

        data_management_plan_results = set()

        for word in query.split():
            num = int(word) if word.isdecimal() else -1

            data_management_plan_results.update(
                DataManagementPlan.objects.filter(
                    is_object_member(user)
                    & Q(projects__deleted__isnull=True, deleted__isnull=True)
                    & (
                        Q(data_management_plan_id=num)
                        | Q(description__icontains=word)
                        | Q(attributes__value__icontains=word)
                    )
                )
            )

        context["data_management_plan_results"] = data_management_plan_results
        return context


##########################################
#               REST API                 #
##########################################


class DataManagementPlanViewSet(HeliportModelViewSet):
    """Data Management Plan"""

    serializer_class = DataManagementPlanSerializer
    filter_backends = [
        rest_filters.SearchFilter,
        django_filters.rest_framework.DjangoFilterBackend,
    ]
    filterset_fields = ["data_management_plan_id"]
    search_fields = ["attributes__value"]

    def get_queryset(self):
        user = self.request.user.heliportuser
        return DataManagementPlan.objects.filter(
            is_object_member(user) & Q(deleted__isnull=True)
        ).distinct()
