# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
import pathlib

from django.db import models

from heliport.core.models import DigitalObject, LoginInfo
from heliport.core.utils.exceptions import AccessDenied

logger = logging.getLogger(__name__)


class DataSource(DigitalObject):
    PROTOCOL_CHOICES = ["https", "http", "ftp", "ftps", "file", "ssh", "other"]

    data_source_id = models.AutoField(primary_key=True)
    uri = models.CharField(max_length=500, blank=True)
    login = models.ForeignKey(
        LoginInfo, on_delete=models.SET_NULL, null=True, blank=True
    )
    generated_inside = models.ForeignKey(
        "DataSource", null=True, on_delete=models.CASCADE
    )

    @property
    def protocol_path(self):
        result = self.uri.split("://", maxsplit=1)
        while len(result) < 2:
            result.append("")
        return result

    @property
    def protocol(self):
        protocol, uri = self.protocol_path
        return protocol

    @property
    def path(self):
        protocol, uri = self.protocol_path
        return uri

    @property
    def browser_supported_uri(self):
        return self.protocol in {"http", "https", "ftp"}

    def sub_path(self, relative_path_str=None):
        """
        This function is deprecated as it is specific to ssh directories.
        The new place is :meth:`heliport.ssh.SSHDirectory.sub_path`.
        """
        source_path = pathlib.Path(self.path)
        relative_path = pathlib.Path("")
        if relative_path_str:
            relative_path = pathlib.Path(relative_path_str)
            if any(n.startswith(".") for n in relative_path.parts):
                raise AccessDenied("Access of hidden directory")

        path = source_path / relative_path
        if source_path != path and source_path not in path.parents:
            raise AccessDenied("Access of parent directory not allowed")
        return path
