# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import pathlib
from abc import ABC
from datetime import datetime, timedelta
from stat import S_ISDIR
from urllib.parse import quote

from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.http import FileResponse, Http404, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render, reverse
from django.utils import timezone
from django.views.generic import CreateView, DetailView, TemplateView, UpdateView

from heliport.core.digital_object_actions import Action
from heliport.core.digital_object_aspects import DirectoryObj, FileObj
from heliport.core.digital_object_resolution import Resolvable, resolve, url_for
from heliport.core.mixins import (
    HeliportLoginRequiredMixin,
    HeliportObjectMixin,
    HeliportProjectMixin,
)
from heliport.core.models import Project, Vocabulary, assert_relation
from heliport.core.permissions import is_object_member
from heliport.core.utils.context import Context, digital_object_breadcrumbs
from heliport.core.utils.exceptions import UserMessage
from heliport.core.utils.string_tools import format_byte_count, format_time
from heliport.core.views import HeliportModelViewSet

from .models import DataSource
from .serializers import DataSourceSerializer

logger = logging.getLogger(__name__)


class DataSourceView(HeliportProjectMixin, CreateView):
    model = DataSource
    template_name = "data_source/data_source.html"
    fields = ["uri", "description", "label", "login"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context["project"]
        context["vocab"] = Vocabulary()
        user = self.request.user.heliportuser
        context["sources"] = [
            {
                "data_source_id": source.data_source_id,
                "digital_object_id": source.digital_object_id,
                "protocol": source.uri.split(":")[0],
                "label": source.label,
                "description": source.description,
                "uri": source.uri,
                "browser_supported_uri": source.browser_supported_uri,
                "owner": (source.login is None)
                or (self.request.user.heliportuser == source.login.user),
                "actions": source.actions(user, project),
            }
            for source in DataSource.objects.filter(
                projects=project, deleted__isnull=True, generated_inside__isnull=True
            )
        ]
        context["protocols"] = DataSource.PROTOCOL_CHOICES
        context["logins"] = user.logininfo_set.all()
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        data_source_id = request.POST.get("remove")
        if data_source_id:
            data_source = get_object_or_404(DataSource, pk=data_source_id)
            data_source.deleted = timezone.now()
            data_source.save()
            return redirect(reverse("data_source:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        protocol = self.request.POST["protocol"]
        server_adress = self.request.POST["server_adress"]

        if protocol == "http" and not form.instance.uri.startswith("http://"):
            form.instance.uri = "http://" + form.instance.uri
        elif protocol == "https" and not form.instance.uri.startswith("https://"):
            form.instance.uri = "https://" + form.instance.uri
        elif protocol == "file" and not form.instance.uri.startswith("file://"):
            form.instance.uri = "file://" + form.instance.uri
        elif protocol == "ftp":
            if not form.instance.uri.startswith("/") and not server_adress.endswith(
                "/"
            ):
                form.instance.uri = "/" + form.instance.uri
            form.instance.uri = server_adress + form.instance.uri
            if not form.instance.uri.startswith("ftp://"):
                form.instance.uri = "ftp://" + form.instance.uri
        elif protocol == "ftps":
            if not form.instance.uri.startswith("/") and not server_adress.endswith(
                "/"
            ):
                form.instance.uri = "/" + form.instance.uri
            form.instance.uri = server_adress + form.instance.uri
            if not form.instance.uri.startswith("ftps://"):
                form.instance.uri = "ftps://" + form.instance.uri
        elif protocol == "ssh":
            form.instance.uri = "ssh://" + form.instance.uri
        else:
            if "://" not in form.instance.uri:
                form.instance.uri = "://" + form.instance.uri

        form.instance.category_str = settings.DATA_SOURCE_NAMESPACE

        form.instance.save()
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.projects.add(project)
        form.instance.register_handle_if_applicable(self.request)

        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_source:list", kwargs=self.kwargs)


class DataSourceUpdateView(HeliportObjectMixin, UpdateView):
    model = DataSource
    template_name = "data_source/data_source.html"
    fields = ["uri", "description", "label", "login"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context["project"]
        context["vocab"] = Vocabulary()
        user = self.request.user.heliportuser
        context["sources"] = [
            {
                "data_source_id": source.data_source_id,
                "digital_object_id": source.digital_object_id,
                "protocol": source.uri.split(":")[0],
                "label": source.label,
                "description": source.description,
                "uri": source.uri,
                "browser_supported_uri": source.browser_supported_uri,
                "owner": (source.login is None)
                or (self.request.user.heliportuser == source.login.user),
                "actions": source.actions(user, project),
            }
            for source in DataSource.objects.filter(
                projects=project, deleted__isnull=True, generated_inside__isnull=True
            )
        ]
        protocol = self.request.GET.get("protocol", self.object.uri.split("://")[0])

        if protocol in ("ftp", "ftps"):
            context["server_adress"] = self.object.uri.split("://", 1)[1].split("/", 1)[
                0
            ]
            context["uri"] = self.object.uri.split("://", 1)[1].split("/", 1)[1]
        else:
            context["uri"] = self.object.uri.split("://", 1)[1]

        context["protocol"] = protocol
        context["protocols"] = DataSource.PROTOCOL_CHOICES
        context["logins"] = user.logininfo_set.all()
        context["login"] = self.object.login
        context["update"] = True
        return context

    def form_valid(self, form):
        if (self.object.login is None) or (
            self.request.user.heliportuser == self.object.login.user
        ):
            protocol = self.request.POST["protocol"]
            server_adress = self.request.POST["server_adress"]

            if protocol == "http" and not form.instance.uri.startswith("http://"):
                form.instance.uri = "http://" + form.instance.uri
            elif protocol == "https" and not form.instance.uri.startswith("https://"):
                form.instance.uri = "https://" + form.instance.uri
            elif protocol == "file" and not form.instance.uri.startswith("file://"):
                form.instance.uri = "file://" + form.instance.uri
            elif protocol == "ftp":
                if not form.instance.uri.startswith("/") and not server_adress.endswith(
                    "/"
                ):
                    form.instance.uri = "/" + form.instance.uri
                form.instance.uri = server_adress + form.instance.uri
                if not form.instance.uri.startswith("ftp://"):
                    form.instance.uri = "ftp://" + form.instance.uri
            elif protocol == "ftps":
                if not form.instance.uri.startswith("/") and not server_adress.endswith(
                    "/"
                ):
                    form.instance.uri = "/" + form.instance.uri
                form.instance.uri = server_adress + form.instance.uri
                if not form.instance.uri.startswith("ftps://"):
                    form.instance.uri = "ftps://" + form.instance.uri
            elif protocol == "ssh":
                form.instance.uri = "ssh://" + form.instance.uri
            else:
                if "://" not in form.instance.uri:
                    form.instance.uri = "://" + form.instance.uri

        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_source:list", kwargs={"project": self.kwargs["project"]})


class SSHDataSourceOpenView(HeliportObjectMixin, DetailView):
    """
    This is the legacy way of showing a directory or downloading a file that is a
    data source with ssh protocol. The new way is to register the ssh files/directories
    in :class:`heliport.ssh.views.SSHListView` and open it using
    :class:`GeneralDataSourceOpenView`.
    """

    model = DataSource

    def get(self, request, *args, **kwargs):
        """
        respond to http get response

        Check if data source is file directory or directory. Download Files using a
        streaming response (ssh connection is closed when django closes file). Show
        directory using template.
        """
        self.object = self.get_object()
        data_source = self.object
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        login = data_source.login
        source_path = pathlib.Path(data_source.path)
        relative_path = pathlib.Path("")
        if "path" in request.GET:
            relative_path = pathlib.Path(request.GET["path"])
            assert not any(
                n.startswith(".") for n in relative_path.parts
            ), "access of hidden directory"

        path = source_path / relative_path
        assert (
            source_path == path or source_path in path.parents
        ), "access of non sub directory"

        logger.debug(f"opening ssh connection datasource={data_source.pk}")
        ssh_connection = login.build_command_executor()
        error_message = ssh_connection.connect_error_message(
            f' - try reconnecting the "{login.name}" ssh connection'
        )
        if error_message is not None:
            messages.error(request, error_message)
            return redirect("data_source:list", project=project.pk)
        file = ssh_connection.file(str(path))

        file_stat = file.stat()
        if hasattr(file_stat, "st_mode") and S_ISDIR(file_stat.st_mode):
            logger.info(f"opening ssh datasource {data_source.pk} (Directory)")
            context = super().get_context_data()
            dir_content_files = ssh_connection.list_dir(str(path))
            dir_content = [
                {
                    "name": s.filename,
                    "path": relative_path / s.filename,
                    "quoted_path": quote(str(relative_path / s.filename)),
                    "digital_object": data_source_by_path(
                        project, path / s.filename, data_source
                    ),
                    "size": hasattr(s, "st_size") and self.format_size(s.st_size),
                    "mtime": hasattr(s, "st_mtime") and self.format_time(s.st_mtime),
                    "is_dir": hasattr(s, "st_mode") and S_ISDIR(s.st_mode),
                }
                for s in dir_content_files
                if not s.filename.startswith(".")
            ]
            context["dir_content"] = sorted(
                dir_content, key=lambda n: (not n["is_dir"], n["name"].lower())
            )
            data_source_to_remove = DataSource.objects.filter(
                projects=project,
                uri__startswith=f"ssh://{path}/",
                deleted__isnull=True,
                generated_inside=data_source,
            )
            for f in dir_content_files:
                data_source_to_remove = data_source_to_remove.exclude(
                    uri__startswith=f"ssh://{path / f.filename}"
                )
            for data_source in data_source_to_remove:
                if path not in pathlib.Path(data_source.path).parents:
                    continue
                logger.debug(f"removing data source {data_source.uri}")
                data_source.deleted = timezone.now()
                data_source.save()

            parent_paths = []
            for p in reversed(relative_path.parents):
                if p != pathlib.Path():
                    parent_paths.append({"name": p.name, "path": str(p)})

            context["parent_paths"] = parent_paths
            context["directory_name"] = path.name
            context["has_relative_path"] = path != pathlib.Path(data_source.path)
            context["vocab"] = Vocabulary()
            return render(request, "data_source/directory.html", context)

        old_close = file.close

        def new_close():
            """
            file response only adds ``file.close`` to functions that need to be called
            to close things. Therefore, it is not completely unnatural to redefine this.
            """
            old_close()
            ssh_connection.close()
            logger.debug(f"closing ssh connection datasource={data_source.pk}")

        file.close = new_close
        response = FileResponse(file, filename=path.name, as_attachment=True)
        file_stat = file.stat()
        if hasattr(file_stat, "st_size"):
            logger.info(
                f"opening ssh datasource {data_source.pk} ({self.format_size(file_stat.st_size)})"
            )
            response["Content-Length"] = file_stat.st_size
        else:
            logger.info(f"opening ssh datasource {data_source.pk}")
        return response

    @staticmethod
    def format_size(size):
        """format file size in bytes to human-readable units"""
        return format_byte_count(size)

    @staticmethod
    def format_time(mtime):
        """formats absolute time in seconds to human-readable string"""
        return format_time(mtime)


class GeneralDataSourceOpenView(HeliportProjectMixin, DetailView):
    """
    Can handle all :class:`heliport.core.digital_object_aspects.DirectoryObj` and
    :class:`heliport.core.digital_object_aspects.FileObj` instances. They could be
    subclasses of :class:`heliport.core.models.DigitalObject` or just generally
    :class:`heliport.core.digital_object_resolution.Resolvable`.

    Files are presented for download and content directories is visualized.
    """

    model = Project
    pk_url_kwarg = "project"

    def get_digital_object(self, context):
        """Get a :class:`heliport.core.digital_object_interface.GeneralDigitalObject`
        instance from the provided URL query parameters using typical resolution.
        """
        obj = resolve(self.request.GET, context)
        if obj is None:
            raise UserMessage(f"Object not found: {self.request.GET}")
        return obj

    def get_directory_response(self, obj: DirectoryObj, exit_context: Context):
        """construct context and render template to show content of directory"""
        render_context = self.get_context_data()
        start_time = datetime.now()
        parts = []
        for part in obj.as_directory(exit_context).get_parts():
            part = part.as_digital_object_if_exists(exit_context)
            self.extend_by_file_info(part, exit_context, start_time)
            parts.append(part)

        render_context["contents"] = sorted(
            parts, key=lambda p: (isinstance(p, FileObj), str(p).casefold())
        )
        render_context["directory"] = obj
        render_context["digital_object_breadcrumbs"] = digital_object_breadcrumbs(
            self.request.session, self.request.GET, obj, self.request.get_full_path()
        )
        return render(
            self.request, "data_source/general_directory.html", render_context
        )

    def extend_by_file_info(self, path_obj, context, start_time):
        """
        Set ``info`` attribute on ``path_obj`` to dict returned by
        :meth:`heliport.core.digital_object_aspects.FileOrDirectory.get_file_info`.

        Does nothing if current time is already more than three seconds after
        ``start_time`` because some ``get_file_info`` implementations might be slow.
        """
        timeout = timedelta(seconds=3)
        if datetime.now() > start_time + timeout:
            path_obj.info = {}
            return

        if isinstance(path_obj, DirectoryObj):
            path_obj.info = path_obj.as_directory(context).get_file_info()
        elif isinstance(path_obj, FileObj):
            path_obj.info = path_obj.as_file(context).get_file_info()

    def get_file_response(self, file_obj: FileObj, context):
        """construct a streaming response for downloading the file"""
        file = file_obj.as_file(context)
        old_close = file.close

        def new_close():
            """
            Close file and context.
            There is no known documented way to set
            `_resource_closers <https://github.com/django/django/blob/26a395f27d86bbf65f330851c8136c33694ac867/django/http/response.py#L570>`_
            otherwise.
            """
            old_close()
            context.close()

        file.open()
        file.close = new_close
        response = FileResponse(
            file,
            filename=str(file_obj),
            as_attachment=True,
            content_type=file.mimetype(),
        )
        size = file.size()
        if size is not None:
            response["Content-Length"] = size

        # prevent that context is closed when exiting with block in get method
        # instead it is made sure that context is closed using new_close function
        context.activate_manual_closing()
        return response

    def get(self, request, *args, **kwargs):
        """respond to http get (this is called by django)"""
        project = self.get_object()
        self.object = project
        with Context(getattr(request.user, "heliportuser", None), project) as context:
            try:
                obj = self.get_digital_object(context)
                if isinstance(obj, DirectoryObj):
                    return self.get_directory_response(obj, context)
                elif isinstance(obj, FileObj):
                    return self.get_file_response(obj, context)
                else:
                    raise UserMessage(
                        f"{obj} did not implement the file or directory interface"
                    )
            except UserMessage as e:
                messages.error(request, e.message)
                raise Http404(e.message) from e


def data_source_by_path(project, path, parent_data_source):
    uri = f"ssh://{path}"
    return (
        DataSource.objects.filter(
            uri=uri,
            login=parent_data_source.login,
            projects=project,
            deleted__isnull=True,
        )
        .order_by("pk")
        .first()
    )


class SubDataSourceCreateView(HeliportObjectMixin, DetailView):
    model = DataSource

    def post(self, request, *args, **kwargs):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        data_source = self.get_object()
        assert data_source.uri.startswith("ssh://")
        login = data_source.login
        path = data_source.sub_path(request.GET.get("path"))
        existing_data_source = data_source_by_path(project, path, data_source)
        if existing_data_source is None:
            new_data_source = DataSource(
                login=login,
                uri=f"ssh://{path}",
                label=path.name,
                generated_inside=data_source,
            )
            new_data_source.category_str = settings.DATA_SOURCE_NAMESPACE
            new_data_source.save()
            new_data_source.projects.add(*data_source.projects.all())
        else:
            new_data_source = existing_data_source
        assert_relation(data_source, Vocabulary().has_part, new_data_source)
        return JsonResponse(
            {
                "pk": new_data_source.pk,
                "digital_object_id": new_data_source.digital_object_id,
                "uri": new_data_source.uri,
                "label": new_data_source.label,
            }
        )


class SearchView(HeliportLoginRequiredMixin, TemplateView):
    template_name = "data_source/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get("q", "")
        context["query"] = query

        data_source_results = set()

        for word in query.split():
            num = int(word) if word.isdecimal() else -1

            data_source_results.update(
                DataSource.objects.filter(
                    is_object_member(user)
                    & Q(projects__deleted__isnull=True, deleted__isnull=True)
                    & (
                        Q(data_source_id=num)
                        | Q(uri__icontains=word)
                        | Q(description__icontains=word)
                        | Q(label__icontains=word)
                        | Q(login__machine__icontains=word)
                        | Q(login__username__icontains=word)
                        | Q(projects__label__icontains=word)
                    )
                )
            )

        context["data_source_results"] = data_source_results
        return context


##########################################
#               REST API                 #
##########################################


class DataSourceViewSet(HeliportModelViewSet):
    """Data Source"""

    serializer_class = DataSourceSerializer
    filterset_fields = ["data_source_id", "label", "uri", "persistent_id", "projects"]

    def get_queryset(self):
        user = self.request.user.heliportuser
        return DataSource.objects.filter(
            is_object_member(user) & Q(deleted__isnull=True)
        ).distinct()


##########################################
#               Actions                  #
##########################################


class OpenAction(Action, ABC):
    """
    Base class for ``Action`` that opens digital objects using
    :class:`GeneralDataSourceOpenView`
    """

    @property
    def link(self) -> str:
        """
        construct a link to :class:`GeneralDataSourceOpenView` specific to current
        project and digital object
        """
        if isinstance(self.obj, Resolvable):
            return url_for("data_source:open", self.obj, project=self.project.pk)
        return "#"

    def applicable(self) -> bool:
        """
        Check applicability by testing if :meth:`link` is successful on this instance.

        This should be overwritten in subclasses to be more specific.
        """
        if hasattr(self.obj, "unsupported") and self.obj.unsupported:
            return False
        return self.link != "#"


class OpenDirectory(OpenAction):
    """
    ``Action`` that opens some ``DirectoryObj`` by showing a preview (see base class).
    """

    #: The name shown to user
    name = "Open"
    #: The font awesome icon possibly shown to user
    icon = "fa-folder-open"

    @property
    def applicable(self) -> bool:
        """
        this action is applicable for
        :class:`heliport.core.digital_object_aspects.DirectoryObj` instances.
        """
        return isinstance(self.obj, DirectoryObj) and super().applicable()


class DownloadFile(OpenAction):
    """``Action`` that downloads some ``FileObj`` (see base class)."""

    #: The name shown to user
    name = "Download"
    #: The font awesome icon possibly used for this action
    icon = "fa-file"

    @property
    def applicable(self) -> bool:
        """
        this action is applicable for
        :class:`heliport.core.digital_object_aspects.FileObj` instances.
        """
        return isinstance(self.obj, FileObj) and super().applicable()
