# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json

from django.contrib.auth.models import User
from django.shortcuts import reverse
from django.test import TestCase

from heliport.core.models import (
    DigitalObject,
    DigitalObjectRelation,
    HeliportUser,
    Project,
)


class DigitalObjectTests(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username="testuser")
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.client.force_login(self.auth_user)
        self.project = Project.objects.create(owner=self.user)
        self.p1 = DigitalObject.objects.create(label="123yyy_predicate")

    def test_list(self):
        obj_project = DigitalObject.objects.create(
            label="123xxx_project", description="a description"
        )
        obj_relation = DigitalObject.objects.create(label="123xxx_relation")
        obj_user = DigitalObject.objects.create(label="123xxx_user")
        obj_none = DigitalObject.objects.create(label="123xxx_none")
        obj_project.projects.add(self.project)
        r = DigitalObjectRelation(
            subject=self.project, predicate=self.p1, object=obj_relation
        )
        r.save()
        obj_user.owner = self.user
        obj_user.save()

        response = self.client.get(
            reverse("digital_objects:list", kwargs={"project": self.project.pk})
        )
        self.assertEqual(200, response.status_code)
        self.assertContains(response, obj_project.label)
        self.assertContains(response, obj_relation.label)
        self.assertContains(response, self.p1.label)
        self.assertNotContains(response, obj_none.label)


class SearchAndAPITest(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username="testuser")
        self.user, is_new = HeliportUser.objects.get_or_create(auth_user=self.auth_user)
        self.project, is_new = Project.objects.get_or_create(owner=self.user)
        self.client.force_login(self.auth_user)

    def test_access_search(self):
        response = self.client.get(reverse("digital_objects:search"))
        self.assertEqual(200, response.status_code)

    def test_findable(self):
        d = DigitalObject.objects.create(label="DIGITAL_OBJECKKKT_123")
        d.projects.add(self.project)
        response = self.client.get(f'{reverse("digital_objects:search")}?q=OBJECKKKT')
        self.assertContains(response, "DIGITAL_OBJECKKKT_123")

    def test_api(self):
        d = DigitalObject.objects.create(description="OBJ654", label="a label")
        d.projects.add(self.project)
        response = self.client.get(reverse("digital_objects:api_root"))
        self.assertEqual(200, response.status_code)
        urls = json.loads(response.content.decode())
        response = self.client.get(urls["digital_objects"])
        self.assertEqual(200, response.status_code)
        self.assertContains(response, "OBJ654")
