# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
import logging
from typing import Union

import requests

from heliport.core.user_logic.user_information import LdapUserInformationManager

logger = logging.getLogger(__name__)


class UserQueries:
    @staticmethod
    @abc.abstractmethod
    def create_user():
        pass

    @staticmethod
    @abc.abstractmethod
    def get_user_by_gate_id(gate_id):
        pass

    @staticmethod
    @abc.abstractmethod
    def get_user_by_ldap_id(ldap_id):
        pass

    @staticmethod
    @abc.abstractmethod
    def get_user_by_ldap_id_check_gate_duplication(ldap_id, gate_id):
        pass

    @staticmethod
    @abc.abstractmethod
    def update_user(user, display_name, ldap_id, gate_id=None):
        pass


class User:
    @abc.abstractmethod
    def check_new(self, user_queries):
        pass

    @abc.abstractmethod
    def save_to(self, user_queries):
        pass

    @abc.abstractmethod
    def get_db_user_from(self, user_queries):
        pass

    @abc.abstractmethod
    def download(self, session: requests.Session):
        pass

    @staticmethod
    @abc.abstractmethod
    def extract_users(project):
        pass

    @classmethod
    def collect_into_set(cls, gate_projects):
        result = set()

        for project in gate_projects:
            result.update(cls.extract_users(project))

        return result

    @classmethod
    def import_users(
        cls,
        gate_projects,
        user_queries,
        session: Union[requests.Session, LdapUserInformationManager],
    ):
        logger.info(f"Import new users of class {cls.__name__}")
        users = cls.collect_into_set(gate_projects)
        new_users = [user for user in users if user.check_new(user_queries)]
        for i, user in enumerate(new_users):
            user.download(session)
            user.save_to(user_queries)
            if i % 50 == 0:
                logger.debug(f"{i} / {len(new_users)} users imported")
        logger.info(f"{len(new_users)} users imported of {len(users)} total users")
