# SPDX-FileCopyrightText: 2023 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import django_filters
from django.conf import settings
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.views.generic import TemplateView
from rest_framework import filters as rest_filters

from heliport.core.mixins import HeliportLoginRequiredMixin
from heliport.core.models import LoginInfo
from heliport.core.permissions import is_object_member
from heliport.core.utils.form_description import FormDescription
from heliport.core.views import HeliportModelViewSet, HeliportObjectListView
from heliport.smb_files.models import DBSMBDirectory, DBSMBFile
from heliport.smb_files.serializers import DirectorySerializer, FileSerializer


class SMBView(HeliportObjectListView):
    """View for showing and editing SMB shares"""

    model = DBSMBDirectory
    category = settings.SMB_DIRECTORY_NAMESPACE
    list_url = "smb_files:list"
    update_url = "smb_files:update"
    list_heading = "SMB Shares"
    create_heading = "Add a SMB Share"

    def register_form_columns(self, form: FormDescription):
        """select which (and how) attributes should be included in edit form"""
        form.add_column("Server Name", "server_name", {"required": True})

        form.add_column("Share Name", "share_name", {"required": True})
        form.add_column(
            "Login",
            "login",
            {
                "widget": "login",
                "type": LoginInfo.LoginTypes.USER_PASSWORD,
                "required": True,
            },
        )
        form.add_column(
            "Domain",
            "domain",
            {
                "placeholder": "can usually be left blank",
                "help_text": "The network domain. "
                "On windows, it is known as the workgroup. "
                "Usually, it is safe to leave this parameter blank.",
            },
        )
        form.add_column("Path", "path_str", {"placeholder": "without share name"})

    def set_update_attributes(self, obj, form_fields, obj_saved=False):
        """get login from pk of html form results and set label"""
        if not obj_saved:
            form_fields["login"] = get_object_or_404(LoginInfo, pk=form_fields["login"])
            obj.label = form_fields["path_str"] or form_fields["share_name"]

        super().set_update_attributes(obj, form_fields, obj_saved)


class SearchView(HeliportLoginRequiredMixin, TemplateView):
    """implement HELIPORT search"""

    template_name = "smb_files/search.html"

    def get_context_data(self, **kwargs):
        """do database query, called by ``TemplateView``"""
        context = super().get_context_data(**kwargs)
        query = self.request.GET.get("q", "")
        context["query"] = query

        smb_file_results = set()
        smb_directory_results = set()

        for word in query.split():
            num = int(word) if word.isdecimal() else -1

            query = Q(deleted__isnull=True) & (
                Q(pk=num)
                | Q(description__icontains=word)
                | Q(attributes__value__icontains=word)
                | Q(path_str__icontains=word)
                | Q(share_name__icontains=word)
                | Q(server_name__icontains=word)
                | Q(domain__icontains=word)
                | Q(login__name__icontains=word)
            )

            smb_file_results.update(DBSMBFile.objects.filter(query))
            smb_directory_results.update(DBSMBDirectory.objects.filter(query))

        context["smb_directory_results"] = smb_directory_results
        context["smb_file_results"] = smb_file_results
        return context


##########################################
#               REST API                 #
##########################################


class FileViewSet(HeliportModelViewSet):
    """SMB Files"""

    serializer_class = FileSerializer
    filter_backends = [
        rest_filters.SearchFilter,
        django_filters.rest_framework.DjangoFilterBackend,
    ]
    filterset_fields = [
        "file_id",
        "persistent_id",
        "digital_object_id",
        "share_name",
        "server_name",
        "path_str",
    ]
    search_fields = ["attributes__value", "path_str", "server_name", "share_name"]

    def get_queryset(self):
        """get files of user that are not deleted"""
        user = self.request.user.heliportuser
        return DBSMBFile.objects.filter(
            is_object_member(user) & Q(deleted__isnull=True)
        ).distinct()


class DirectoryViewSet(HeliportModelViewSet):
    """SMB Directories"""

    serializer_class = DirectorySerializer
    filter_backends = [
        rest_filters.SearchFilter,
        django_filters.rest_framework.DjangoFilterBackend,
    ]
    filterset_fields = [
        "directory_id",
        "persistent_id",
        "digital_object_id",
        "share_name",
        "server_name",
        "path_str",
    ]
    search_fields = ["attributes__value", "path_str", "server_name", "share_name"]

    def get_queryset(self):
        """get directories of user that are not deleted"""
        user = self.request.user.heliportuser
        return DBSMBDirectory.objects.filter(
            is_object_member(user) & Q(deleted__isnull=True)
        ).distinct()
